"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module contains the Schema Maker View Service client.
"""

import asyncio
from typing import Annotated, Literal, Optional

from pydantic import Field

from pyegeria.core._server_client import ServerClient
from pyegeria.models import (
    NewElementRequestBody,
    DeleteElementRequestBody,
    UpdateElementRequestBody,
    SearchStringRequestBody,
    GetRequestBody,
    ReferenceableProperties,
)
from pyegeria.view.output_formatter import (
    populate_common_columns,
    overlay_additional_values,
)
from pyegeria.core.utils import dynamic_catch


class SchemaTypeProperties(ReferenceableProperties):
    class_: Annotated[Literal["SchemaTypeProperties"], Field(alias="class")]
    is_deprecated: Optional[bool] = None
    author: Optional[str] = None
    usage: Optional[str] = None
    encoding_standard: Optional[str] = None
    namespace: Optional[str] = None


class SchemaAttributeProperties(ReferenceableProperties):
    class_: Annotated[Literal["SchemaAttributeProperties"], Field(alias="class")]
    element_position: Optional[int] = None
    min_cardinality: Optional[int] = None
    max_cardinality: Optional[int] = None
    allows_duplicate_values: Optional[bool] = None
    is_ordered_values: Optional[bool] = None
    default_value_override: Optional[str] = None
    anchor_guid: Optional[str] = None


class SchemaMaker(ServerClient):
    """
    Client for the Schema Maker View Service.

    The Schema Maker View Service provides methods to manage schema types and schema attributes.

    Attributes
    ----------
    view_server : str
        The name of the View Server to use.
    platform_url : str
        URL of the server platform to connect to.
    user_id : str
        The identity of the user calling the method.
    user_pwd : str
        The password associated with the user_id. Defaults to None.
    """

    def __init__(
        self,
        view_server: str,
        platform_url: str,
        user_id: str,
        user_pwd: Optional[str] = None,
        token: Optional[str] = None,
    ):
        super().__init__(view_server, platform_url, user_id, user_pwd, token)
        self.view_server = view_server
        self.platform_url = platform_url
        self.user_id = user_id
        self.user_pwd = user_pwd
        self.url_marker = "schema-maker"

    def _extract_schema_properties(self, element: dict, columns_struct: dict) -> dict:
        col_data = populate_common_columns(element, columns_struct)
        props = element.get("properties", {})
        overlay_additional_values(col_data, props)
        return col_data

    def _generate_schema_output(
        self,
        elements: dict | list[dict],
        filter_string: Optional[str] = None,
        element_type_name: Optional[str] = None,
        output_format: str = "DICT",
        report_spec: dict | str | None = None,
        **kwargs,
    ) -> str | list[dict]:
        return self._generate_formatted_output(
            elements=elements,
            query_string=filter_string,
            element_type_name=element_type_name or "SchemaType",
            output_format=output_format,
            report_spec=report_spec,
            extract_properties_func=self._extract_schema_properties,
            **kwargs,
        )

    # Schema Types

    @dynamic_catch
    async def _async_create_schema_type(self, body: dict | NewElementRequestBody) -> str:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-types"
        return await self._async_create_element_body_request(url, "SchemaTypeProperties", body)

    def create_schema_type(self, body: dict | NewElementRequestBody) -> str:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._async_create_schema_type(body))

    @dynamic_catch
    async def _async_update_schema_type(
        self, schema_type_guid: str, body: dict | UpdateElementRequestBody
    ) -> None:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-types/{schema_type_guid}/update"
        await self._async_update_element_body_request(url, body)

    def update_schema_type(self, schema_type_guid: str, body: dict | UpdateElementRequestBody) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_update_schema_type(schema_type_guid, body))

    @dynamic_catch
    async def _async_delete_schema_type(
        self, schema_type_guid: str, body: dict | DeleteElementRequestBody
    ) -> None:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-types/{schema_type_guid}/delete"
        await self._async_delete_element_body_request(url, body)

    def delete_schema_type(self, schema_type_guid: str, body: dict | DeleteElementRequestBody) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_delete_schema_type(schema_type_guid, body))

    # Schema Attributes

    @dynamic_catch
    async def _async_create_schema_attribute(self, body: dict | NewElementRequestBody) -> str:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-attributes"
        return await self._async_create_element_body_request(url, "SchemaAttributeProperties", body)

    def create_schema_attribute(self, body: dict | NewElementRequestBody) -> str:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._async_create_schema_attribute(body))

    @dynamic_catch
    async def _async_update_schema_attribute(
        self, schema_attribute_guid: str, body: dict | UpdateElementRequestBody
    ) -> None:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-attributes/{schema_attribute_guid}/update"
        await self._async_update_element_body_request(url, body)

    def update_schema_attribute(
        self, schema_attribute_guid: str, body: dict | UpdateElementRequestBody
    ) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_update_schema_attribute(schema_attribute_guid, body))

    @dynamic_catch
    async def _async_delete_schema_attribute(
        self, schema_attribute_guid: str, body: dict | DeleteElementRequestBody
    ) -> None:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-attributes/{schema_attribute_guid}/delete"
        await self._async_delete_element_body_request(url, body)

    def delete_schema_attribute(
        self, schema_attribute_guid: str, body: dict | DeleteElementRequestBody
    ) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._async_delete_schema_attribute(schema_attribute_guid, body))

    @dynamic_catch
    async def _async_find_schema_types(
        self,
        search_string: str = "*",
        body: Optional[dict | SearchStringRequestBody] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = "JSON",
        report_spec: str | dict = "Referenceable",
        **kwargs
    ) -> list | str:
        """ Retrieve the list of schema type metadata elements that contain the search string. Async Version.

        Parameters
        ----------
        search_string: str
            Search string to match against - None or '*' indicate match against all schema types.
        starts_with : bool, [default=True], optional
            Starts with the supplied string.
        ends_with : bool, [default=False], optional
            Ends with the supplied string
        ignore_case : bool, [default=False], optional
            Ignore case when searching
        anchor_domain: str, optional
            The anchor domain to search in.
        metadata_element_type: str, optional
            The type of metadata element to search for.
        metadata_element_subtypes: list[str], optional
            The subtypes of metadata element to search for.
        skip_relationships: list[str], optional
            The types of relationships to skip.
        include_only_relationships: list[str], optional
            The types of relationships to include.
        skip_classified_elements: list[str], optional
            The types of classified elements to skip.
        include_only_classified_elements: list[str], optional
            The types of classified elements to include.
        graph_query_depth: int, [default=3], optional
            The depth of the graph query.
        governance_zone_filter: list[str], optional
            The governance zones to search in.
        as_of_time: str, optional
            The time to search as of.
        effective_time: str, optional
            The effective time to search at.
        relationship_page_size: int, [default=0], optional
            The page size for relationships.
        limit_results_by_status: list[str], optional
            The statuses to limit results by.
        sequencing_order: str, optional
            The order to sequence results by.
        sequencing_property: str, optional
            The property to sequence results by.
        output_format: str, default = "JSON"
            - one of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON"
        report_spec: str | dict , optional, default = "Referenceable"
            - The desired output columns/fields to include.
        start_from: int, [default=0], optional
            When multiple pages of results are available, the page number to start from.
        page_size: int, [default=100]
            The number of items to return in a single page.
        property_names: list[str], optional
            The names of properties to search for.
        body: dict | SearchStringRequestBody, optional, default = None
            - if provided, the search parameters in the body will supercede other attributes, such as "search_string"

        Returns
-------
        List | str

        Output depends on the output format specified.

        Raises
-------

        ValidationError
          If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
          Issues raised in communicating or server side processing.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-types/by-search-string"
        
        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec
        }
        params.update(kwargs)
        
        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}
        
        response = await self._async_find_request(url, _type="SchemaType", _gen_output=self._generate_schema_output,
                                                  **params)

        return response

    @dynamic_catch
    def find_schema_types(
        self,
        search_string: str = "*",
        body: Optional[dict | SearchStringRequestBody] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = "JSON",
        report_spec: str | dict = "Referenceable",
        **kwargs
    ) -> list | str:
        """ Retrieve the list of schema type metadata elements that contain the search string.

        Parameters
        ----------
        search_string: str
            Search string to match against - None or '*' indicate match against all schema types.
        starts_with : bool, [default=True], optional
            Starts with the supplied string.
        ends_with : bool, [default=False], optional
            Ends with the supplied string
        ignore_case : bool, [default=False], optional
            Ignore case when searching
        anchor_domain: str, optional
            The anchor domain to search in.
        metadata_element_type: str, optional
            The type of metadata element to search for.
        metadata_element_subtypes: list[str], optional
            The subtypes of metadata element to search for.
        skip_relationships: list[str], optional
            The types of relationships to skip.
        include_only_relationships: list[str], optional
            The types of relationships to include.
        skip_classified_elements: list[str], optional
            The types of classified elements to skip.
        include_only_classified_elements: list[str], optional
            The types of classified elements to include.
        graph_query_depth: int, [default=3], optional
            The depth of the graph query.
        governance_zone_filter: list[str], optional
            The governance zones to search in.
        as_of_time: str, optional
            The time to search as of.
        effective_time: str, optional
            The effective time to search at.
        relationship_page_size: int, [default=0], optional
            The page size for relationships.
        limit_results_by_status: list[str], optional
            The statuses to limit results by.
        sequencing_order: str, optional
            The order to sequence results by.
        sequencing_property: str, optional
            The property to sequence results by.
        output_format: str, default = "JSON"
            - one of "MD", "LIST", "FORM", "REPORT", "DICT", "MERMAID" or "JSON"
        report_spec: str | dict , optional, default = "Referenceable"
            - The desired output columns/fields to include.
        start_from: int, [default=0], optional
            When multiple pages of results are available, the page number to start from.
        page_size: int, [default=100]
            The number of items to return in a single page.
        property_names: list[str], optional
            The names of properties to search for.
        body: dict | SearchStringRequestBody, optional, default = None
            - if provided, the search parameters in the body will supercede other attributes, such as "search_string"

        Returns
-------
        List | str

        Output depends on the output format specified.

        Raises
-------

        ValidationError
          If the client passes incorrect parameters on the request that don't conform to the data model.
        PyegeriaException
          Issues raised in communicating or server side processing.
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_schema_types(
                search_string=search_string,
                body=body,
                starts_with=starts_with,
                ends_with=ends_with,
                ignore_case=ignore_case,
                start_from=start_from,
                page_size=page_size,
                output_format=output_format,
                report_spec=report_spec,
                **kwargs
            )
        )

    @dynamic_catch
    async def _async_get_schema_type_by_guid(
        self,
        schema_type_guid: str,
        element_type: str = "SchemaType",
        body: Optional[dict | GetRequestBody] = None,
        output_format: str = "JSON",
        report_spec: str | dict = "SchemaTypes",
    ) -> dict | str:
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/schema-maker/schema-types/{schema_type_guid}/retrieve"
        return await self._async_get_guid_request(
            url,
            _type=element_type,
            _gen_output=self._generate_schema_output,
            output_format=output_format,
            report_spec=report_spec,
            body=body,
        )

    def get_schema_type_by_guid(
        self,
        schema_type_guid: str,
        element_type: str = "SchemaType",
        body: Optional[dict | GetRequestBody] = None,
        output_format: str = "JSON",
        report_spec: str | dict = "SchemaTypes",
    ) -> dict | str:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_schema_type_by_guid(
                schema_type_guid, element_type, body, output_format, report_spec
            )
        )



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.
"""

"""
This common file is used to set some global values and enumerations used by the overall package. 

"""
from enum import Enum
from typing import Any, Type

def resolve_enum(enum_class: Type[Any], value: str | int) -> int | None:
    """
    Resolves a string or integer to its corresponding Egeria enum integer value.
    Example: resolve_enum(ContentStatus, 'Draft') -> 0
    """
    if value is None or value == "":
        return None
        
    if isinstance(value, int):
        return value
        
    v = str(value).upper()
    try:
        # Normalized case-insensitive search through enum members
        # Replace separators with spaces and collapse multiple spaces
        v_clean = ' '.join(v.replace('_', ' ').replace('-', ' ').split())
        for member in enum_class:
            m_name = ' '.join(member.name.upper().replace('_', ' ').replace('-', ' ').split())
            if v_clean == m_name:
                return member.value

        # 3. Check if it's already an integer string
        if v.isdigit():
            return int(v)
        return None
    except (KeyError, ValueError):
        return None

is_debug = False
disable_ssl_warnings = True
enable_ssl_check = False
max_paging_size = 500
default_time_out = 30
DEBUG_LEVEL = "quiet"
COMMENT_TYPES = (
    "STANDARD_COMMENT",
    "ANSWER",
    "OTHER",
    "QUESTION",
    "SUGGESTION",
    "USAGE_EXPERIENCE",
    "REQUIREMENT"
)
star_ratings = (
    "FIVE_STARS",
    "FOUR_STARS",
    "NO_RECOMMENDATION",
    "ONE_STAR",
    "THREE_STARS",
    "TWO_STARS",
)
# class GovernanceDomains(Enum):
#     All= 0
#     Data= 1
#     Privacy= 2
#     Security= 3
#     IT Infrastructure= 4
#     Software Development= 5
#     Corporate= 6
#     Asset Management= 7
#     Other= 99
GovernanceDomains = Enum('GovernanceDomains', [('Unclassified', 0),
                                               ('Data', 1),
                                               ('Privacy',2),
                                               ('Security',3),
                                               ('IT Infrastructure',4),
                                               ('Software Development',5),
                                               ('Corporate',6),
                                               ('Asset Management',7),
                                               ('Other',99)])

# Default status values for fallback if dynamic fetching fails
CONTENT_STATUS = ["DRAFT", "PREPARED", "PROPOSED", "APPROVED", "REJECTED", "ACTIVE", "DEPRECATED", "OTHER"]
DEPLOYMENT_STATUS = ["PROPOSED", "UNDER_DEVELOPMENT", "DEVELOPMENT_COMPLETE", "APPROVED_FOR_DEPLOYMENT", 
                     "REJECTED_FOR_DEPLOYMENT", "STANDBY", "ACTIVE", "DISABLED", "FAILED", "OTHER"]
ACTIVITY_STATUS = ["REQUESTED", "APPROVED", "WAITING", "ACTIVATING", "IN_PROGRESS", "PAUSED", "FOR_INFO", 
                   "COMPLETED", "INVALID", "IGNORED", "FAILED", "CANCELLED", "ABANDONED", "OTHER"]
MEMBERSHIP_STATUS = ["UNKNOWN", "DISCOVERED", "PROPOSED", "IMPORTED", "VALIDATED", "DEPRECATED", "OBSOLETE", "OTHER"]

RELATIONSHIP_TYPES = ["RelatedTerm", "Synonym", "Antonym", "PreferredTerm", "ReplacementTerm", "Translation", "IsA", "ValidValue"]

# RELATIONSHIP_TYPES_MAP = {rel_type: i for i, rel_type in enumerate(RELATIONSHIP_TYPES)}

TEMPLATE_GUIDS: dict[str, str] = {}
INTEGRATION_GUIDS: dict[str, str] = {}

NO_ELEMENTS_FOUND = "No elements found"
NO_ASSETS_FOUND = "No assets found"
NO_SERVERS_FOUND = "No servers found"
NO_CATALOGS_FOUND = "No catalogs found"
NO_GLOSSARIES_FOUND = "No glossaries found"
NO_TERMS_FOUND = "No terms found"
NO_CATEGORIES_FOUND = "No categories found"
NO_ELEMENT_FOUND = "No element found"
NO_PROJECTS_FOUND = "No projects found"
NO_COLLECTION_FOUND = "No collection found"
NO_GUID_RETURNED = "No guid returned"
NO_MEMBERS_FOUND = "No members found"

MERMAID_GRAPHS: list[str] = ["anchorMermaidGraph", "informationSupplyChainMermaidGraph","fieldLevelLineageGraph",
                  "actionMermaidGraph", "localLineageGraph", "edgeMermaidGraph", "iscImplementationMermaidGraph",
                  "specificationMermaidGraph", "solutionBlueprintMermaidGraph","mermaidGraph",
                  "solutionSubcomponentMermaidGraph", "governanceActionProcessMermaidGraph"]
MERMAID_GRAPH_TITLES = ["Anchor Mermaid Graph", "Information Supply Chain Mermaid Graph","Field Level Lineage Graph",
                  "Action Mermaid Graph", "Local Lineage Graph", "Edge Mermaid Graph", "ISC Implementation Mermaid Graph",
                  "Specification Mermaid Graph", "Solution Blueprint Mermaid Graph","Mermaid Graph",
                  "Solution Subcomponent Mermaid Graph","Governance Action Process Mermaid Graph" ]
TERM_STATUS = ["DRAFT", "PREPARED","PROPOSED","APPROVED", "REJECTED", "ACTIVE", "DEPRECATED", "DELETED", "OTHER"]
REPORT_ACRONYMS = ["GUID", "URL", "ID", "QN", "API", "UI"]


"""
Standardized extraction logic for Dr.Egeria v2.
"""
import re
from dataclasses import dataclass, field
from typing import List, Dict, Optional, Any
from loguru import logger

# Standard Dr.Egeria / OMAG verbs
STANDARD_VERBS = {
    "Create", "Update", "Delete", "Link", "Attach", "Add", 
    "Unlink", "Detach", "Remove", "Display", "Run", "View", 
    "Search", "Find", "Provenance", "Validate", "Process"
}

@dataclass
class DrECommand:
    verb: str
    object_type: str
    attributes: Dict[str, str] = field(default_factory=dict)
    raw_block: str = ""
    start_line: int = 0
    end_line: int = 0
    is_command: bool = True

class UniversalExtractor:
    """
    Extracts Dr.Egeria commands from various text formats:
    - Standard Markdown (# Verb Object)
    - Jupyter Notebooks (Markdown cells)
    - LLM prompts (Headless or Markdown)
    """
    def __init__(self, text: str):
        self.text = text
        # Regex to find command start: Match # <Verb> <Object> or a headless line starting with a verb
        # Accounts for horizontal rules or whitespace that might be the actual first characters of the block text
        verbs_pattern = "|".join(STANDARD_VERBS)
        self.cmd_header_rx = re.compile(
            rf"^(?:\s*_+|[-]+\s*\n)?(?:#\s+)?(?P<verb>{verbs_pattern})\s+(?P<object>[^#\n_]+)", 
            re.MULTILINE | re.IGNORECASE
        )

    def extract_commands(self) -> List[DrECommand]:
        blocks = self._split_into_blocks()
        commands = []
        
        for block_text, start_line in blocks:
            match = self.cmd_header_rx.search(block_text)
            if match:
                verb = match.group("verb").strip().capitalize()
                obj = match.group("object").strip()
                
                # Cleanup object name from trailing markdown artifacts
                obj = re.sub(r'[\s_*-]+$', '', obj)
                
                attributes = self._extract_attributes_from_block(block_text)
                
                commands.append(DrECommand(
                    verb=verb,
                    object_type=obj,
                    attributes=attributes,
                    raw_block=block_text,
                    start_line=start_line,
                    end_line=start_line + block_text.count('\n'),
                    is_command=True
                ))
            else:
                # Preservation: include non-command blocks as-is
                commands.append(DrECommand(
                    verb="",
                    object_type="",
                    attributes={},
                    raw_block=block_text,
                    start_line=start_line,
                    end_line=start_line + block_text.count('\n'),
                    is_command=False
                ))
        return commands

    def _split_into_blocks(self) -> List[tuple[str, int]]:
        """Split text into potential command blocks based on H1 headers or horizontal rules."""
        # Using a lookahead to split by '#' at start of line OR horizontal rules
        # Also handles headless blocks by checking if the start is a command
        lines = self.text.splitlines()
        blocks = []
        current_block = []
        start_line = 1
        
        for i, line in enumerate(lines):
            # Check for block boundaries: H1 header or horizontal rule
            if (line.startswith("# ") or re.match(r'^\s*___+\s*$', line) or re.match(r'^\s*---+\s*$', line)):
                if current_block:
                    blocks.append(("\n".join(current_block), start_line))
                current_block = [line]
                start_line = i + 1
            else:
                current_block.append(line)
        
        if current_block:
            blocks.append(("\n".join(current_block), start_line))
            
        return blocks

    def _extract_attributes_from_block(self, block: str) -> Dict[str, str]:
        """Extracts ## attributes from a block of text."""
        attributes = {}
        # Match ## Header until next ## or end of block
        attr_rx = re.compile(r"^##\s+(?P<label>[^#\n]+)\n(?P<value>(?:(?!^##).)*)", re.MULTILINE | re.DOTALL)
        
        for match in attr_rx.finditer(block):
            label = match.group("label").strip()
            value = match.group("value").strip()
            
            # Clean up the value (remove provenance lines, excessive newlines)
            filtered_lines = [
                line for line in value.splitlines() 
                if not line.lstrip().startswith(">") and not re.match(r'^\s*_+\s*$', line)
            ]
            attributes[label] = "\n".join(filtered_lines).strip()
            
        return attributes



"""
This file contains term-related object_action functions for processing Egeria Markdown
"""
import json
import os
from typing import Optional

from inflect import engine
from jsonschema import ValidationError
from loguru import logger
from rich import print
from rich.console import Console
from rich.markdown import Markdown

from md_processing.md_processing_utils.common_md_proc_utils import (
    parse_upsert_command,
    parse_view_command,
    render_command_table,
    render_exception_table,
)
from md_processing.md_processing_utils.common_md_utils import (set_gov_prop_body,
                                                               set_update_body, set_create_body,
                                                               set_peer_gov_def_request_body, set_rel_prop_body,
                                                               set_rel_request_body,
                                                               ALL_GOVERNANCE_DEFINITIONS, set_delete_request_body)
from md_processing.md_processing_utils.extraction_utils import (extract_command_plus, update_a_command)
from md_processing.md_processing_utils.md_processing_constants import (load_commands)
from pyegeria import DEBUG_LEVEL, body_slimmer, PyegeriaException, GovernanceDomains, print_basic_exception, print_validation_error
from pyegeria.egeria_tech_client import EgeriaTech
from pyegeria.core.utils import make_format_set_name_from_type

GERIA_METADATA_STORE = os.environ.get("EGERIA_METADATA_STORE", "active-metadata-store")
EGERIA_KAFKA_ENDPOINT = os.environ.get("KAFKA_ENDPOINT", "localhost:9092")
EGERIA_PLATFORM_URL = os.environ.get("EGERIA_PLATFORM_URL", "https://localhost:9443")
EGERIA_VIEW_SERVER = os.environ.get("EGERIA_VIEW_SERVER", "view-server")
EGERIA_VIEW_SERVER_URL = os.environ.get("EGERIA_VIEW_SERVER_URL", "https://localhost:9443")
EGERIA_INTEGRATION_DAEMON = os.environ.get("EGERIA_INTEGRATION_DAEMON", "integration-daemon")
EGERIA_INTEGRATION_DAEMON_URL = os.environ.get("EGERIA_INTEGRATION_DAEMON_URL", "https://localhost:9443")
EGERIA_ADMIN_USER = os.environ.get("ADMIN_USER", "garygeeke")
EGERIA_ADMIN_PASSWORD = os.environ.get("ADMIN_PASSWORD", "secret")
EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_WIDTH = os.environ.get("EGERIA_WIDTH", 220)
EGERIA_JUPYTER = os.environ.get("EGERIA_JUPYTER", False)
EGERIA_HOME_GLOSSARY_GUID = os.environ.get("EGERIA_HOME_GLOSSARY_GUID", None)
EGERIA_GLOSSARY_PATH = os.environ.get("EGERIA_GLOSSARY_PATH", None)
EGERIA_ROOT_PATH = os.environ.get("EGERIA_ROOT_PATH", "../../")
EGERIA_INBOX_PATH = os.environ.get("EGERIA_INBOX_PATH", "md_processing/dr_egeria_inbox")
EGERIA_OUTBOX_PATH = os.environ.get("EGERIA_OUTBOX_PATH", "md_processing/dr_egeria_outbox")

GENERAL_GOVERNANCE_DEFINITIONS = ["Business Imperative", "Regulation Article", "Threat", "Governance Principle",
                                  "Governance Obligation", "Governance Approach", "Governance Processing Purpose"]

GOVERNANCE_CONTROLS = ["Governance Rule", "Service Level Objective", "Governance Process",
                       "Governance Responsibility", "Governance Procedure", "Security Access Control"]

load_commands('commands.json')
debug_level = DEBUG_LEVEL

console = Console(width=int(250))


#
# Helper functions for the governance officer commands
#

@logger.catch
def sync_gov_rel_elements(egeria_client: EgeriaTech, object_action: str, object_type: str, guid: str,
                          qualified_name: str, attributes: dict):
    # TODO: when the next release is available, I should be able to more easily get the asis elements - so will
    # TODO: hold off on implementing replace all
    try:
        merge_update = attributes.get("Merge Update", {}).get("value", True)
        to_be_supports_policies = attributes.get("Supports Policies", {}).get("guid_list", None)
        to_be_governance_drivers = attributes.get("Governance Drivers", {}).get("guid_list", None)

        if merge_update or object_action == "Create":
            if to_be_supports_policies:
                for policy in to_be_supports_policies:
                    egeria_client.attach_supporting_definitions(policy, "GovernanceImplementation", guid)
                    print(f"Added `{policy}` to `{guid}`")
            elif to_be_governance_drivers:
                for policy in to_be_governance_drivers:
                    egeria_client.attach_supporting_definitions(policy, "GovernanceResponse", guid)
                    print(f"Added `{policy}` to `{guid}`")
    except Exception as ex:
        print(ex)


@logger.catch
def process_gov_definition_upsert_command(egeria_client: EgeriaTech, txt: str,
                                          directive: str = "display") -> Optional[str]:
    """
    Processes a data specification create or update object_action by extracting key attributes.

    :param txt: Command details in text format.
    :param directive: An optional string indicating the directive - display, validate, or process.
    :return: A string summarizing the outcome, or None if no processing is needed.
    """
    try:
        # Extract basic command information
        command, object_type, object_action = extract_command_plus(txt)
        if not (command and object_type and object_action):
            logger.error("Failed to parse `command`, `object_type`, or `object_action` from input text.")
            return None
        if object_type not in ALL_GOVERNANCE_DEFINITIONS:
            logger.error(f"Invalid object type: {object_type}")
            return None
        # Log command details
        print(Markdown(f"# {command}\n"))

        # Parse command attributes
        parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)
        if not parsed_output or 'valid' not in parsed_output:
            logger.error(f"Unable to parse command properly: {txt}")
            return None

        valid = parsed_output['valid']
        exists = parsed_output['exists']
        qualified_name = parsed_output.get('qualified_name')
        display_name = parsed_output.get('attributes', {}).get('Display Name', {}).get('value')
        guid = parsed_output.get('guid')
        version_identifier = parsed_output.get('attributes',{}).get('Version Identifier', {}).get('value')
        print(Markdown(parsed_output.get('display', '')))
        logger.debug(json.dumps(parsed_output, indent=4))

        if directive == "display":
            logger.info("Directive set to display. No processing required.")
            return None

        elif directive == "validate":
            if valid:
                pass
            else:
                logger.error(f"Validation failed for `{command}`.")
            return valid

        elif directive == "process":
            output_set = make_format_set_name_from_type(object_type)

            if object_action == "Update":
                if not guid:
                    msg = (f"The `{object_type}` '{display_name}' does not yet exist.\n The result document has been "
                           f"updated to change `Update` to `Create` in processed output\n{'-' * 80}\n ")
                    logger.error(msg)
                    # print(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)

                if not valid:
                    logger.error("Invalid data for update action.")
                    return None

                # Proceed with the update
                update_body = set_update_body(object_type, parsed_output['attributes'])
                update_body['properties'] = set_gov_prop_body(object_type, qualified_name, parsed_output['attributes'])
                update_body = body_slimmer(update_body)
                egeria_client.update_governance_definition(guid, update_body)
                if status := parsed_output['attributes'].get('Status', {}).get('value', None):
                    egeria_client.update_element_status(guid, status)
                logger.success(f"Updated {object_type} `{display_name}` with GUID {guid}")
                return egeria_client.get_governance_definition_by_guid(guid, output_format='MD',
                                                                       report_spec=output_set)

            elif object_action == "Create":
                if valid is False and exists:
                    msg = (f"Failed to create `{object_type}` named `{display_name}` because it already exists.\n "
                           f"Result document has been updated to change `Create` to `Update` in processed o"
                           f"utput\n{'-' * 80}\n")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif valid is False:
                    msg = f" invalid data? - Correct and try again\n\n___"
                    logger.error(msg)
                    return None
                else:
                    if qualified_name is None:
                        qualified_name = egeria_client.__create_qualified_name__(object_type,display_name,
                                                                                 local_qualifier=None, version_identifier=version_identifier)
                    create_body = set_create_body(object_type, parsed_output['attributes'])
                    create_body['properties'] = set_gov_prop_body(object_type, qualified_name,
                                                                  parsed_output['attributes'])
                    guid = egeria_client.create_governance_definition(body_slimmer(create_body))
                    if guid:
                        logger.success(f"Created {object_type} `{display_name}` with GUID {guid}")
                        return egeria_client.get_governance_definition_by_guid(guid, output_format='MD',
                                                                               report_spec=output_set)
                    else:
                        logger.error(f"Failed to create {object_type} `{display_name}`.")
                        return None

        else:
            logger.error(f"Unsupported directive: {directive}")
            return None

    except PyegeriaException as e:
        logger.error(f"PyegeriaException occurred: {e}")
        render_exception_table(command, "process", e)
        return None

    except Exception as e:
        logger.exception(f"Unexpected error: {e}")
        print("Unexepected error occurred:\n {e}")
        return None


@logger.catch
def process_gov_def_link_detach_command(egeria_client: EgeriaTech, txt: str,
                                        directive: str = "display") -> Optional[str]:
    """
    Processes a link or unlink command to associate or break up peer governance definitions.

    :param txt: A string representing the input cell to be processed for
        extracting blueprint-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_view_command(egeria_client, object_type, object_action, txt, directive)

    render_command_table(parsed_output, directive)

    logger.debug(json.dumps(parsed_output, indent=4))

    attributes = parsed_output['attributes']

    definition1 = attributes.get('Definition 1', {}).get('guid', None)
    definition2 = attributes.get('Definition 2', {}).get('guid', None)
    label = attributes.get('Link Label', {}).get('value', None)
    description = attributes.get('Description', {}).get('value', None)

    valid = parsed_output['valid']
    exists = definition1 is not None and definition2 is not None

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            pass
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":
        # Initialize the inflect engine
        inflector = engine()

        # Split the object_type into words
        object_type_parts = object_type.split()

        # Determine which word to singularize (third word if present, otherwise fallback to last word)
        if len(object_type_parts) >= 3:
            word_to_singularize = object_type_parts[2]
        else:
            word_to_singularize = object_type_parts[-1]  # Fallback to last word

        # Singularize the selected word
        singular_word = inflector.singular_noun(word_to_singularize) or word_to_singularize

        # Construct gov_peer_relationship_type
        gov_peer_relationship_type = f"Governance{singular_word}Link"

    try:
        if object_action in ["Detach", "Unlink", "Remove"]:
            if not exists:
                msg = (f" Link `{label}` does not exist! Updating result document with Link "
                       f"object_action\n")
                logger.error(msg)
                out = parsed_output['display'].replace('Link', 'Detach', 1)
                return out
            elif not valid:
                return None
            else:
                print(Markdown(
                    f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))
                body = set_delete_request_body(object_type, attributes)

            egeria_client.detach_peer_definitions(definition1, gov_peer_relationship_type,
                                                  definition2, body)

            logger.success(f"===> Detached segment with {label} from `{definition1}`to {definition2}\n")
            out = parsed_output['display'].replace('Unlink', 'Link', 1)

            return (out)


        elif object_action in ["Link", "Attach", "Add"]:
            if valid is False and exists:
                msg = (f"-->  Link called `{label}` already exists and result document updated changing "
                       f"`Link` to `Detach` in processed output\n")
                logger.error(msg)

            elif valid is False:
                msg = f"==>{object_type} Link with label `{label}` is not valid and can't be created"
                logger.error(msg)
                return

            else:
                body = set_peer_gov_def_request_body(object_type, attributes)
                egeria_client.link_peer_definitions(definition1, gov_peer_relationship_type,
                                                    definition2, body)
                msg = f"==>Created {object_type} link named `{label}`\n"
                logger.success(msg)
                out = parsed_output['display'].replace('Link', 'Detach', 1)
                return out

    except ValidationError as e:
        render_exception_table(command, "process", e)
        logger.error(f"Validation Error performing {command}: {e}")
        return None
    except PyegeriaException as e:
        render_exception_table(command, "process", e)
        logger.error(f"PyegeriaException occurred: {e}")
        return None

    except Exception as e:
        logger.error(f"Error performing {command}: {e}")
        return None
    else:
        return None


@logger.catch
def process_supporting_gov_def_link_detach_command(egeria_client: EgeriaTech, txt: str,
                                                   directive: str = "display") -> Optional[str]:
    """
    Processes a link or unlink command to associate or break up peer governance definitions.

    :param txt: A string representing the input cell to be processed for
        extracting blueprint-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_view_command(egeria_client, object_type, object_action, txt, directive)

    render_command_table(parsed_output, directive)

    logger.debug(json.dumps(parsed_output, indent=4))

    attributes = parsed_output['attributes']

    definition1 = attributes.get('Definition 1', {}).get('guid', None)
    definition2 = attributes.get('Definition 2', {}).get('guid', None)
    label = attributes.get('Link Label', {}).get('value', None)
    description = attributes.get('Description', {}).get('value', None)

    valid = parsed_output['valid']
    exists = definition1 is not None and definition2 is not None

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            pass
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":

        # Construct gov_peer_relationship_type
        relationship_type_name = object_type.replace(' ', '')
        print(f"relationship_type_name: {relationship_type_name}")
    try:
        if object_action in ["Detach", "Unlink", "Remove"]:
            if not exists:
                msg = (f" Link `{label}` does not exist! Updating result document with Link "
                       f"object_action\n")
                logger.error(msg)
                out = parsed_output['display'].replace('Link', 'Detach', 1)
                return out
            elif not valid:
                return None
            else:
                print(Markdown(
                    f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))
                body = set_delete_request_body(object_type, attributes)

            egeria_client.detach_supporting_definitions(definition1, relationship_type_name,
                                                        definition2, body)

            logger.success(f"===> Detached segment with {label} from `{definition1}`to {definition2}\n")
            out = parsed_output['display'].replace('Unlink', 'Link', 1)

            return (out)


        elif object_action in ["Link", "Attach", "Add"]:
            if valid is False and exists:
                msg = (f"-->  Link called `{label}` already exists and result document updated changing "
                       f"`Link` to `Detach` in processed output\n")
                logger.error(msg)

            elif valid is False:
                msg = f"==>{object_type} Link with label `{label}` is not valid and can't be created"
                logger.error(msg)
                return None

            else:
                body_prop = {
                    "class": "SupportingDefinitionProperties",
                    "rationale": attributes.get('Rationale', {}).get('value', None),
                    "effectiveFrom": attributes.get('Effective From', {}).get('value', None),
                    "effectiveTo": attributes.get('Effective To', {}).get('value', None),
                }

                body = set_peer_gov_def_request_body(object_type, attributes)
                body['properties'] = body_prop
                egeria_client.attach_supporting_definitions(definition1, relationship_type_name,
                                                            definition2, body)
                msg = f"==>Created {object_type} link named `{label}`\n"
                logger.success(msg)
                out = parsed_output['display'].replace('Link', 'Detach', 1)
                return out

    except ValidationError as e:
        render_exception_table(command, "process", e)
        logger.error(f"Validation Error performing {command}: {e}")
        return None
    except PyegeriaException as e:
        render_exception_table(command, "process", e)
        logger.error(f"PyegeriaException occurred: {e}")
        return None

    except Exception as e:
        logger.error(f"Error performing {command}: {e}")
        return None
    else:
        return None



@logger.catch
def process_governed_by_link_detach_command(egeria_client: EgeriaTech, txt: str,
                                                   directive: str = "display") -> Optional[str]:
    """
    Processes a link or unlink command to associate or break up peer governance definitions.

    :param txt: A string representing the input cell to be processed for
        extracting blueprint-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_view_command(egeria_client, object_type, object_action, txt, directive)

    render_command_table(parsed_output, directive)

    logger.debug(json.dumps(parsed_output, indent=4))

    attributes = parsed_output['attributes']

    element_guid = attributes.get('Referenceable', {}).get('guid', None)
    definition_guid = attributes.get('Governance Definition', {}).get('guid', None)
    label = attributes.get('Link Label', {}).get('value', None)
    description = attributes.get('Description', {}).get('value', None)

    valid = parsed_output['valid']
    exists = element_guid is not None and definition_guid is not None

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            pass
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":


        try:
            if object_action in ["Detach", "Unlink", "Remove"]:
                if not exists:
                    msg = (f" Link `{label}` does not exist! Updating result document with Link "
                           f"object_action\n")
                    logger.error(msg)
                    out = parsed_output['display'].replace('Link', 'Detach', 1)
                    return out
                elif not valid:
                    return None
                else:
                    print(Markdown(
                        f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))
                    body = set_delete_request_body(object_type, attributes)

                egeria_client.detach_governed_by_definitions(element_guid, definition_guid, body)

                logger.success(f"===> Detached segment with {label} from `{definition1}`to {definition2}\n")
                out = parsed_output['display'].replace('Unlink', 'Link', 1)

                return (out)


            elif object_action in ["Link", "Attach", "Add"]:
                if valid is False and exists:
                    msg = (f"-->  Link called `{label}` already exists and result document updated changing "
                           f"`Link` to `Detach` in processed output\n")
                    logger.error(msg)

                elif valid is False:
                    msg = f"==>{object_type} Link with label `{label}` is not valid and can't be created"
                    logger.error(msg)
                    return

                else:
                    body_prop = set_rel_prop_body(object_type, attributes)

                    body = set_rel_request_body(object_type, attributes)
                    body['properties'] = body_prop
                    body = body_slimmer(body)
                    egeria_client.attach_governed_by_definition(element_guid, definition_guid, body)
                    msg = f"==>Created {object_type} link named `{label}`\n"
                    logger.success(msg)
                    out = parsed_output['display'].replace('Link', 'Detach', 1)
                    return out

        except ValidationError as e:
            render_exception_table(command, "process", e)
            logger.error(f"Validation Error performing {command}: {e}")
            return None
        except PyegeriaException as e:
            render_exception_table(command, "process", e)
            logger.error(f"PyegeriaException occurred: {e}")
            return None

        except Exception as e:
            logger.error(f"Error performing {command}: {e}")
            return None
    else:
        return None



@logger.catch
def process_gov_def_context_command(egeria_client: EgeriaTech, txt: str,
                                    directive: str = "display") -> Optional[str]:
    """
    Retrieves the context graph for a governance definition.

    :param txt: A string representing the input cell to be processed for
        extracting term-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_view_command(egeria_client, object_type, object_action, txt, directive)

    valid = parsed_output['valid']
    exists = parsed_output['exists']
    qualified_name = parsed_output.get('qualified_name', None)
    display_name = parsed_output.get('display_name', None)
    guid = parsed_output.get('guid', None)

    render_command_table(parsed_output, directive)

    attr = parsed_output.get('attributes', {})

    output_format = attr.get('Output Format', {}).get('value', 'LIST')
    detailed = attr.get('Detailed', {}).get('value', False)

    if directive == "display":
        return None
    elif directive == "validate":
        if valid:
            pass
        else:
            msg = f"Validation failed for object_action `{command}`\n"
            logger.error(msg)
        return valid

    elif directive == "process":
        try:
            if not valid:  # First validate the command before we process it
                msg = f"Validation failed for {object_action} `{object_type}`\n"
                logger.error(msg)
                return None

            list_md = f"\n# `{object_type}` with Qualified Name: `{qualified_name}`\n\n"
            body = set_update_body(object_type, attr)

            struct = egeria_client.get_gov_def_in_context(guid, body=body, output_format=output_format)
            if output_format.upper() == "DICT":
                list_md += f"```\n{json.dumps(struct, indent=4)}\n```\n"
            else:
                list_md += struct
            logger.info(f"Wrote `{object_type}` graph for : `{qualified_name}`")

            return list_md

        except Exception as e:
            logger.error(f"Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

 OMAG Server configuration functions.  These functions add definitions to an OMAG server's configuration document.
 This class encompasses the full set of configuration methods.
"""

import json
from typing import Any, Optional
from pyegeria.core._exceptions import (
    PyegeriaInvalidParameterException,
    PyegeriaAPIException,
)
from pyegeria.core._validators import validate_name, validate_url

from pyegeria.omvs.core_omag_server_config import CoreServerConfig


class FullServerConfig(CoreServerConfig):
    """
    This class represents a client for configuring the OMAG Server.

    Parameters
    ----------
    server_name : str
        The name of the server to connect to.
    platform_url : str
        The URL of the platform where the server is running.
    user_id : str
        The user ID for authentication.
    user_pwd : str, optional
        The password for authentication (default: None).
    verify_flag : bool, optional
        Flag to enable/disable SSL certificate verification (default: enable_ssl_check).

    """

    def __init__(
        self,
        server_name: str,
        platform_url: str,
        user_id: str,
        user_pwd: Optional[str] = None,
    ):
        self.full_config_command_root: str
        CoreServerConfig.__init__(self, server_name, platform_url, user_id, user_pwd)
        self.full_config_command_root = (
            self.platform_url + "/open-metadata/admin-services"
        )

    def get_access_services_topic_names(
        self, access_service_name: str, server_name: str = None
    ) -> list[str]:
        """Retrieve the topic names for this access service.
        Parameters
        ----------
        access_service_name : str
            The name of the access service.

        server_name : str, optional
            The name of the server. If not provided, the server name of the instance is used.

        Returns
        -------
        list[str]
            A list of topic names associated with the specified access service.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(access_service_name)
        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/access-services/"
            + access_service_name
            + "/topic-names"
        )
        response = self.make_request("GET", url)

        return response.json()  # todo fix

    def get_all_access_services_topic_names(self, server_name: str = None) -> list:
        """Retrieve the topic names for all access services.
        Parameters
        ----------
        server_name : str, optional
            The name of the server. If not provided, the default server name will be used.

        Returns
        -------
        list
            The JSON response containing the access services topic names.

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/access-services/topic-names"
        )
        response = self.make_request("GET", url)

        return response.json()  # todo fix

    def set_access_services_configuration(
        self, access_services_body: str, server_name: str = None
    ) -> None:
        """Set up the configuration for selected open metadata access services (OMASs).
        This overrides the current configured values.

        Parameters
        ----------
        access_services_body: str
            The body of the access services configuration. This should be a string representation of the desired
            configuration.

        server_name: str, optional
            The name of the server. If not provided, the default server will be used.

        Returns
        -------
        None

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/access-services/configuration"
        )
        self.make_request("POST", url, access_services_body)
        return

    def override_access_service_in_topic_name(
        self, access_service_name: str, new_topic_name: str, server_name: str = None
    ) -> None:
        """override the in topic for the access service"""
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/access-services/"
            + access_service_name
            + "/topic-names/in-topic"
        )
        self.make_request("POST", url, new_topic_name)
        return

    def override_access_service_out_topic_name(
        self, access_service_name: str, new_topic_name: str, server_name: str = None
    ) -> None:
        """override the out topic for the access service"""
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/access-services/"
            + access_service_name
            + "/topic-names/out-topic"
        )
        self.make_request("POST", url, new_topic_name)
        return

    def set_audit_log_destinations(
        self, audit_dest_body: str, server_name: str = None
    ) -> None:
        """Sets the audit log destinations for a server

        /open-metadata/admin-services/users/{userId}/servers/{serverName}/audit-log-destinations

        Parameters
        ----------
        audit_dest_body : str
            A JSON document describing the audit destinations for this server.

        server_name : str
            Name of the server to update.

        Returns
        -------
        None

        Raises
        ------

        PyegeriaInvalidParameterException
            If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
            The principle specified by the user_id does not have authorization for the requested action

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/audit-log-destinations"
        )
        self.make_request("POST", url, audit_dest_body)
        return

    def update_audit_log_destination(
        self, connection_name: str, audit_dest_body: str, server_name: str = None
    ) -> None:
        """Update an audit log destination that is identified with the supplied destination name
            with the supplied connection object.
        Parameters
        ----------
        connection_name : str
            The name of the connection for the audit log destination.
        audit_dest_body : str
            The body of the audit log destination.
        server_name : str, optional
            The name of the server. If not provided, the server name associated with the instance will be used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response status code is not 200 and the related HTTP code is not 200.

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/audit-log-destinations/connection/"
            + connection_name
        )
        self.make_request("POST", url, audit_dest_body)
        return

    def add_audit_log_destination(
        self, connection_body: str, server_name: str = None
    ) -> None:
        """Adds an audit log destination to a server.

        /open-metadata/admin-services/users/{userId}/servers/{serverName}/audit-log-destinations/connection

        Parameters
        ----------
        connection_body : str
            JSON string containing the connection properties for the audit log destination.
        server_name : str
            Name of the server to update.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException:
            If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action
        ConfigurationErrorException:
            Raised when configuration parameters passed on earlier calls turn out to be
            invalid or make the new call invalid.

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/audit-log-destinations/connection"
        )
        self.make_request("POST", url, connection_body)
        return

    def set_cohort_config(
        self, cohort_name: str, cohort_config_body: str, server_name: str = None
    ) -> None:
        """Set up the configuration properties for a cohort. This may reconfigure an existing cohort or
            create a cohort. Use setCohortMode to delete a cohort.
        Parameters
        ----------
        cohort_name : str
            The name of the cohort for which the configuration is being set.
        cohort_config_body : str
            The body of the cohort configuration.
        server_name : str, optional
            The name of the server. If not provided, the method will use the server name that is already set.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the `cohort_config_body` is None.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)
        if cohort_config_body is None:
            raise PyegeriaInvalidParameterException(
                None,
                context={"caller_method": "set_cohort_config"},
                additional_info={"reason": "cohort_config_body is None"},
            )

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/configuration"
        )
        self.make_request("POST", url, cohort_config_body)
        return

    def clear_cohort_configuration(
        self, cohort_name: str, server_name: str = None
    ) -> None:
        """Retrieves the stored configurations for a server
        Parameters
        ----------
        cohort_name : str

        server_name : str, optional

        Returns
        -------

        Raises
        ------
         PyegeriaInvalidParameterException
             If the client passes incorrect parameters on the request - such as bad URLs or invalid values
         PyegeriaAPIException
             Raised by the server when an issue arises in processing a valid request
         NotAuthorizedException
             The principle specified by the user_id does not have authorization for the requested action

        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/cohorts/{cohort_name}"
        validate_name(cohort_name)

        self.make_request("DELETE", url)

    def get_dedicated_cohort_topic_names(
        self, cohort_name: str, server_name: str = None
    ) -> list:
        """Retrieve the current topic name for the cohort. This call can only be made once the cohort is set up with
            add_cohort_registration().

        Parameters
        ----------
        cohort_name : str
            The name of the cohort for which to retrieve the dedicated topic names.

        server_name : str, optional
            The name of the server on which the cohort resides. If not provided, the default server name will be used.

        Returns
        -------
        list
            A list of topic names that are dedicated to the specified cohort.

        """
        if server_name is None:
            server_name = self.server_name

        validate_name(cohort_name)

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/dedicated-topic-names"
        )
        response = self.make_request("POST", url)
        return response.json().get("topicNames")

    def get_cohort_topic_name(self, cohort_name: str, server_name: str = None) -> str:
        """Retrieve the current topic name for the cohort. This call can only be made once the
            cohort is set up with add_cohort_registration().
        Parameters
        ----------
        cohort_name : str
            The name of the cohort for which to retrieve the topic name.
        server_name : str, optional
            The name of the server on which the cohort is located. If not provided, the default server name will be used.

        Returns
        -------
        str
            The topic name associated with the specified cohort.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the request to retrieve the topic name fails or returns an invalid parameter code.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)

        url = f"{self.full_config_command_root}/servers/{server_name}/cohorts/{cohort_name}/topic-name"
        response = self.make_request("GET", url)
        return response.json().get("topicName")

    def override_cohort_topic_name(
        self, cohort_name: str, topic_override: str, server_name: str = None
    ) -> None:
        """Override the current name for the single topic for the cohort. This call can only be made once the
            cohort is set up with add_cohort_registration().
        Parameters
        ----------
        cohort_name : str
            The name of the cohort for which the topic name is to be overridden.

        topic_override : str
            The new topic name to override the original topic name.

        server_name : str, optional
            The name of the server. If not provided, the default server name will be used.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)
        validate_name(topic_override)

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/topic-name-override"
        )
        self.make_request("POST", url, topic_override)
        return

    def override_instances_cohort_topic_name(
        self, cohort_name: str, topic_override: str, server_name: str = None
    ) -> None:
        """Override the current name for the "instances" topic for the cohort. This call can only be made once
            the cohort is set up with add_cohort_registration().
        Parameters
        ----------
        cohort_name : str
            The name of the cohort to override the topic name for.

        topic_override : str
            The new topic name to override for the cohort.

        server_name : str, optional
            The name of the server where the cohort is located. If not specified,
            the default server name will be used.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)
        validate_name(topic_override)

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/topic-name-override/instances"
        )
        self.make_request("POST", url, topic_override)
        return

    def override_registration_cohort_topic_name(
        self, cohort_name: str, topic_override: str, server_name: str = None
    ) -> None:
        """Override the current name for the registration topic for the cohort. This call can only be made once
            the cohort is set up with add_cohort_registration().
        Parameters
        ----------
        cohort_name : str
            Name of the cohort to override the topic name for.
        topic_override : str
            The overriding topic name for the cohort.
        server_name : str, optional
            Name of the server. If not provided, the default server name will be used.

        Returns
        -------
        None
            This method does not return any value.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)
        validate_name(topic_override)

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/topic-name-override/registration"
        )
        self.make_request("POST", url, topic_override)
        return

    def override_types_cohort_topic_name(
        self, cohort_name: str, topic_override: str, server_name: str = None
    ) -> None:
        """Override the current name for the "types" topic for the cohort. This call can only be made once
            the cohort is set up with add_cohort_registration().
        Parameters
        ----------
        cohort_name : str
            The name of the cohort.
        topic_override : str
            The topic override to be set for the cohort.
        server_name : str, optional
            The name of the server. If not provided, it will use the default server name from the class.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response from the server has a relatedHTTPCode other than 200.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)
        validate_name(topic_override)

        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/cohorts/"
            + cohort_name
            + "/topic-name-override/types"
        )
        self.make_request("POST", url, topic_override)
        return

    def add_full_cohort_registration(
        self, cohort_name: str, topic_structure: str, server_name: str = None
    ) -> None:
        """add a full cohort registration

        Parameters
        ----------
        cohort_name : str
            Name of the cohort to be registered.
        topic_structure : str
            Topic structure for the cohort. This is a string from an enumerated list.
            'Dedicated Cohort Topics', description='The cohort members use three topics to exchange information.
            One for registration requests, one for type validation and one for exchange of instances stored by the
            cohort members.
                This is the preferred and optimal approach
             'Single Topic', description='All asynchronous communication between cohort members is via a single topic.
                This is the original design and may still be used when communicating with back level cohort members.
            'Both Single and Dedicated Topics', description='Both the single cohort topic and the dedicated topics are
                set up and used. This is necessary when the cohort has members with different capabilities.
                 This configuration may cause some events to be processed twice.'
        server_name : str, optional
            Name of the server to which the cohort should be added. Defaults to None.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200 after making the POST request.

        """
        if server_name is None:
            server_name = self.server_name
        validate_name(cohort_name)

        url = f"{self.full_config_command_root}/servers/{server_name}/cohorts/{cohort_name}/topic-structure/{topic_structure}"
        self.make_request("POST", url)

    def set_server_configuration(
        self, config_body: str, server_name: str = None
    ) -> None | str:
        """Sets the configurations for a server
        Parameters
        ----------

        Returns
        -------
        str
            The stored configurations for the given server.

        Raises
        ------
         PyegeriaInvalidParameterException
             If the client passes incorrect parameters on the request - such as bad URLs or invalid values
         PyegeriaAPIException
             Raised by the server when an issue arises in processing a valid request
         NotAuthorizedException
             The principle specified by the user_id does not have authorization for the requested action
         ConfigurationErrorException
             Raised when configuration parameters passed on earlier calls turn out to be
             invalid or make the new call invalid.

        """
        if server_name is None:
            server_name = self.server_name
        url = (
            self.full_config_command_root + "/servers/" + server_name + "/configuration"
        )
        response = self.make_request("POST", url, config_body)
        if response.status_code != 200:
            return str(response.status_code)  # should never get here?

        related_code = response.json().get("relatedHTTPCode")
        if related_code == 200:
            return
        else:
            raise PyegeriaAPIException(response)

    def clear_stored_configuration(self, server_name: str = None) -> None | str:
        """Retrieves the stored configurations for a server
        Parameters
        ----------

        Returns
        -------
        str
            The stored configurations for the given server.

        Raises
        ------
         PyegeriaInvalidParameterException
             If the client passes incorrect parameters on the request - such as bad URLs or invalid values
         PyegeriaAPIException
             Raised by the server when an issue arises in processing a valid request
         NotAuthorizedException
             The principle specified by the user_id does not have authorization for the requested action
         ConfigurationErrorException
             Raised when configuration parameters passed on earlier calls turn out to be
             invalid or make the new call invalid.

        """
        if server_name is None:
            server_name = self.server_name
        url = (
            self.full_config_command_root + "/servers/" + server_name + "/configuration"
        )
        response = self.make_request("DELETE", url)
        if response.status_code != 200:
            return str(response.status_code)  # should never get here?

        related_code = response.json().get("relatedHTTPCode")
        if related_code == 200:
            return
        else:
            raise PyegeriaAPIException(response)

    def deploy_stored_configurations(
        self, target_url_root: str, server_name: str = None
    ) -> None:
        """Push the configuration for the server to another OMAG Server Platform.
        Parameters
        ----------
        target_url_root : str
            The target URL root where the configurations will be deployed.
        server_name : str, optional
            The name of the server to which the configurations will be deployed.
            If not provided, the default server name will be used.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the client passes incorrect parameters on the request — such as bad URLs or invalid values.
        PyegeriaAPIException
            Raised by the server when an issue arises in processing a valid request.
        PyegeriaUnauthorizedException
            The principal specified by the `user_id` does not have authorization for the requested action.

        """
        if server_name is None:
            server_name = self.server_name

        url = f"{self.full_config_command_root}/servers/{server_name}/configuration/deploy"

        self.make_request("POST", url, target_url_root)

    def get_event_bus(self, server_name: str = None) -> dict:
        """Returns the event bus configuration for the specified server

        Parameters
        ----------
        server_name: str, optional
            The name of the server to retrieve the event bus configuration from. If not provided, the default
            server name specified in the class instance will be used.

        Returns
        -------
            The event bus configuration as a JSON dictionary

        Raises
        ------
        PyegeriaInvalidParameterException
            If the client passes incorrect parameters on the request — such as bad URLs or invalid values.
        PyegeriaAPIException
            Raised by the server when an issue arises in processing a valid request.
        PyegeriaUnauthorizedException
            The principal specified by the `user_id` does not have authorization for the requested action.

        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/event-bus"

        response = self.make_request("GET", url)
        return response.json().get("config")

    def set_event_bus_detailed(
        self, connector_provider: str, topic_url_root: str, server_name: str = None
    ) -> None:
        """
        Parameters
        ----------

        connector_provider : str
            Name of the connector provider

        topic_url_root : str
            URL for the topic

        server_name : str, optional
            The name of the server to retrieve the event bus configuration from. If not provided, the default server
            name specified in the class instance will be used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the client passes incorrect parameters on the request — such as bad URLs or invalid values.

        Description
        -----------
        Set up the default event bus for embedding in event-driven connector.
        The resulting connector will be used for example, in the OMRS Topic Connector for each cohort,
        the in and out topics for each Access Service and possibly the local repository's event mapper.
        When the event bus is configured, it is used only on future configuration.
        It does not affect existing configuration.

        """

        if server_name is None:
            server_name = self.server_name

        url = (
            f"{self.full_config_command_root}/servers/{server_name}/event-bus?connectorProvider="
            f"{connector_provider}&topicURLRoot={topic_url_root}"
        )
        self.make_request("POST", url)

    def delete_event_bus(self, server_name: str = None) -> None:
        """Delete the event bus configuration for the given server.

        Parameters
        ----------

        server_name : str, optional
            The name of the server to retrieve the event bus configuration from. If not provided, the default server
            name specified in the class instance will be used.

        Returns
        -------

        Raises
        ------
        PyegeriaInvalidParameterException
            If the client passes incorrect parameters on the request — such as bad URLs or invalid values.

        Description
        -----------
        Delete the current configuration for the event bus. This does not impact that existing configuration for the
        server, only future configuration requests.
        """

        if server_name is None:
            server_name = self.server_name

        url = f"{self.full_config_command_root}/servers/{server_name}/event-bus"
        self.make_request("DELETE", url)

    def set_server_url_root(self, url_root: str, server_name: str = None) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/server-url-root-for-caller"
        body = {"urlRoot": url_root}
        response = self.make_request("POST", url, payload=body)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_max_page_size(self, max_page_size: int, server_name: str = None) -> None:
        """
            Set the maximum page size for a server.

        Parameters
        ----------
        max_page_size : int
            The maximum page size to set.
        server_name : str, optional
            The name of the server for which to set the maximum page size. If not specified, the default server name
            will be used.

        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action


        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/max-page-size?limit={max_page_size}"
        self.make_request("POST", url)

    def set_server_user_id(self, server_user_id: str, server_name: str = None) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/server-user-id?id={server_user_id}"
        response = self.make_request("POST", url)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_server_user_password(
        self, server_user_pwd: str, server_name: str = None
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/server-user-password?password={server_user_pwd}"
        response = self.make_request("POST", url)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_organization_name(self, org_name: str, server_name: str = None) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/organization-name?name={org_name}"
        response = self.make_request("POST", url)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_local_repository_config(
        self, repository_body: str, server_name: str = None
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        url = (
            self.full_config_command_root
            + "/servers/"
            + server_name
            + "/local-repository/configuration"
        )
        response = self.make_request("POST", url, payload=repository_body)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_plug_in_repository_connection(
        self, repository_connection_body: str, server_name: str = None
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/local-repository/mode/plugin-repository/connection"
        self.make_request("POST", url, payload=repository_connection_body)

    def set_plug_in_repository_connection_provider(
        self, repository_connection_provider: str, server_name: str = None
    ) -> None:
        """Set the local repository connection with a user specified connection provider

        Parameters
        ----------
        repository_connection_provider : str
            The name of the connection provider for the plugin repository.

        server_name : str, optional
            The name of the server. If not provided, the default server name will be used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        """
        if server_name is None:
            server_name = self.server_name
        url = (
            f"{self.full_config_command_root}/servers/{server_name}/local-repository/mode/plugin-repository/"
            f"details?connectionProvider={repository_connection_provider}"
        )
        self.make_request("POST", url)

    def set_open_metadata_archives(
        self, archives_list_body: str, server_name: str = None
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/open-metadata-archives"
        response = self.make_request("POST", url, payload=archives_list_body)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_descriptive_server_type(
        self, type_name: str, server_name: str = None
    ) -> None:
        """Set descriptiveServerType for this OMAG server

        Parameters
        ----------
        type_name : str
            The name of the descriptive server type to set.

        server_name : str, optional
            The name of the server for which the descriptive server type is being set. If not provided, the
            default server name associated with the object is used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/server-type?typeName={type_name}"
        self.make_request("POST", url)

    def set_server_description(self, server_desc: str, server_name: str = None) -> None:
        if server_name is None:
            server_name = self.server_name
        url = (
            f"{self.full_config_command_root}/servers/{server_name}/server-description"
        )
        response = self.make_request("POST", url, server_desc)
        related_code = response.json().get("relatedHTTPCode")
        if related_code != 200:
            raise PyegeriaAPIException(response)
        else:
            return

    def set_server_type(self, server_type: str, server_name: str = None) -> None:
        """Sets the server type for the given server

        Parameters
        ----------
        server_type : str
            The type of server to set

        server_name : str, optional
            The name of the server. If None, the default server name will be used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        Notes
        -----

        See https://egeria-project.org/concepts/omag-server/#server-name for details.

        """
        if server_name is None:
            server_name = self.server_name

        url = (
            f"{self.full_config_command_root}/servers/{server_name}/"
            f"server-type?typename={server_type}"
        )
        self.make_request("POST", url)

        def clear_server_type(self, server_name: str = None) -> None:
            """Clears the server type for the given server

            Parameters
            ----------
            server_name : str, optional
                The name of the server. If None, the default server name will be used.

            Returns
            -------
            None

            Raises
            ------
            PyegeriaInvalidParameterException
                If the response code is not 200.
            PyegeriaAPIException:
                Raised by the server when an issue arises in processing a valid request
            NotAuthorizedException:
                The principle specified by the user_id does not have authorization for the requested action

            """
            if server_name is None:
                server_name = self.server_name
            url = f"{self.full_config_command_root}/servers/{server_name}/server-type?typeName="
            self.make_request("POST", url)

    def config_view_service(
        self, service_url_marker: str, view_service_body: dict, server_name: str = None
    ) -> None:
        """Configure a the view service specified by the service_url_marker using the view_service_body.

        Parameters
        ----------
        service_url_marker : str
            The service URL marker. A list can be retrieved through the `list_registered_view_svcs` of the
            `registered_info` module.

        view_service_body : dict
            The body of the view service request.

        server_name : str, optional
            The name of the server. If not provided, the value of `self.view_server` will be used.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        Notes
        -----

        Details on view services configuration can be found at:
         https://egeria-project.org/guides/admin/servers/configuring-the-view-services/?h=view#integration-view-services

        """

        if server_name is None:
            server_name = self.server_name
        validate_name(service_url_marker)

        url = f"{self.full_config_command_root}/servers/{server_name}/view-services/{service_url_marker}"
        self.make_request("POST", url, view_service_body)

    # todo - this may not be used anymore - old
    def set_view_svcs_config(
        self, view_svcs_config_body: dict, server_name: str = None
    ) -> None:
        """Set up the configuration for all the open metadata integration groups. This overrides the current values.

        Parameters
        ----------
        view_svcs_config_body : dict
            The configuration body for the view services.

        server_name : str, optional
            The name of the server. If not provided, the default server name will be used.

        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        Notes
        -----

        Details on view services configuration can be found at:
         https://egeria-project.org/guides/admin/servers/configuring-the-view-services/?h=view#integration-view-services

        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/view-services/configuration"
        self.make_request("POST", url, view_svcs_config_body)

    def set_integration_groups_config(
        self, integration_groups_config_body: dict, server_name: str = None
    ) -> None:
        """Set up the configuration for all the open metadata integration groups. This overrides the current values.
        Parameters
        ----------
        integration_groups_config_body : dict
            The configuration body for the integration groups.

        server_name : str, optional
            The name of the server. If not provided, the default server name will be used.

        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
            If the response code is not 200.
        PyegeriaAPIException:
            Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException:
            The principle specified by the user_id does not have authorization for the requested action

        Notes
        -----

        Details on integration group configuration can be found at:
         https://egeria-project.org/concepts/integration-group/

        body format is:
        [
          {
            "integrationGroupQualifiedName": "string",
            "omagserverPlatformRootURL": "string",
            "omagserverName": "string"
          }
        ]
        """
        if server_name is None:
            server_name = self.server_name
        url = f"{self.full_config_command_root}/servers/{server_name}/integration-groups/configuration/all"
        self.make_request("POST", url, integration_groups_config_body)

    def set_engine_host_services(self):
        # handles an array of engine list - pass in JSON

        pass

    def set_integration_daemon_services_configuration(self):
        # pass a json list
        pass

    def config_integration_service(
        self,
        remote_omag_server: str,
        remote_omag_platform_url: str,
        service_url_marker: str,
        integration_service_options: dict,
        connector_configs: list,
        server_name: Optional[str] = None,
    ) -> None:
        if server_name is None:
            server_name = self.server_name

        if not isinstance(connector_configs, list):
            exc_msg = " ==> connector_configs must be a list of dictionaries"
            raise Exception(exc_msg)

        validate_name(remote_omag_server)
        validate_url(remote_omag_platform_url)

        validate_name(service_url_marker)

        request_body = {
            "class": "IntegrationServiceRequestBody",
            "omagserverPlatformRootURL": remote_omag_platform_url,
            "omagserverName": remote_omag_server,
            "integrationServiceOptions": integration_service_options,
            "integrationConnectorConfigs": connector_configs,
        }

        url = f"{self.full_config_command_root}/servers/{server_name}/integration-services/{service_url_marker}"
        # print(f"URL is : {url}")
        # print(f"body is : \n{json.dumps(request_body, indent=4)}")

        self.make_request("POST", url, request_body)
        return

    def config_all_integration_services(
        self,
        remote_omag_server: str,
        remote_omag_platform_url: str,
        integration_service_options: dict,
        connector_configs: dict,
        server_name: Optional[str] = None,
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        validate_name(remote_omag_server)
        validate_url(remote_omag_platform_url)

        request_body = {
            "IntegrationConnectorConfigs": [
                {
                    "class": "IntegrationServiceRequestBody",
                    "omagserverPlatformRootURL": remote_omag_platform_url,
                    "omagserverName": remote_omag_server,
                    "integrationServiceOptions": integration_service_options,
                    "integrationConnectorConfigs": connector_configs,
                }
            ]
        }

        url = f"{self.full_config_command_root}/servers/{server_name}/integration-services"
        print(f"URL is : {url}")
        print(f"body is : \n{json.dumps(request_body, indent=4)}")

        self.make_request("POST", url, request_body)

    def clear_integration_service(
        self, service_url_marker: str, server_name: str = None
    ) -> None:
        if server_name is None:
            server_name = self.server_name
        validate_name(service_url_marker)

        url = f"{self.full_config_command_root}/servers/{server_name}/integration-services/{service_url_marker}"
        self.make_request("DELETE", url)

    def get_integration_service_config(
        self, service_url_marker: str, server_name: str = None
    ) -> dict | str:
        if server_name is None:
            server_name = self.server_name
        validate_name(service_url_marker)
        url = f"{self.full_config_command_root}/servers/{server_name}/integration-services/{service_url_marker}/configuration"
        response = self.make_request("GET", url)
        return response.json().get("config", "No configuration found")

    def get_integration_services_configs(self, server_name: str = None) -> dict | str:
        if server_name is None:
            server_name = self.server_name

        url = f"{self.full_config_command_root}/servers/{server_name}/integration-services/configuration"
        response = self.make_request("GET", url)
        return response.json().get("services", "No configuration found")

    def set_lineage_warehouse_services(self, body: dict, lineage_server: str) -> None:
        url = f"{self.full_config_command_root}/servers/{lineage_server}/lineage-warehouse/configuration"
        self.make_request("POST", url, body)

    def remove_lineage_warehouse_services(self, lineage_server: str = None) -> None:
        url = f"{self.full_config_command_root}/servers/{lineage_server}/lineage-warehouse/configuration"
        self.make_request("DELETE", url)


if __name__ == "__main__":
    print("Main-Full OMAG Server Config")



"""
This is an ongoing experiment in parsing and playing with Freddie docs
"""
import os
import sys
import click
from loguru import logger
from rich.console import Console
from rich.prompt import Prompt

from pyegeria.core._exceptions import PyegeriaException
from pyegeria.core.config import settings
from md_processing.dr_egeria import process_md_file_v2
import asyncio

# Configure logging
log_format = "{time} | {level} | {function} | {line} | {message} | {extra}"
logger.remove()
logger.add(sys.stderr, level="INFO", format=log_format, colorize=True)
logger.add("debug_log.log", rotation="1 day", retention="1 week", compression="zip", level="TRACE", format=log_format,
           colorize=True)

# Load configuration from config/config.json with environment variable overrides
app_config = settings.Environment

# Get configuration values with environment variable fallbacks
EGERIA_VIEW_SERVER = os.environ.get("EGERIA_VIEW_SERVER", app_config.egeria_view_server)
EGERIA_VIEW_SERVER_URL = os.environ.get("EGERIA_VIEW_SERVER_URL", app_config.egeria_view_server_url)
# User credentials are only from environment variables or command line (not stored in config for security)
EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_WIDTH = int(os.environ.get("EGERIA_WIDTH", app_config.console_width or 220))
EGERIA_JUPYTER = os.environ.get("EGERIA_JUPYTER", str(app_config.egeria_jupyter)).lower() in ("true", "1", "yes")

console = Console(width=EGERIA_WIDTH)

@click.command("process_markdown_file", help="Process a markdown file and return the output as a string.")
@click.option("--input-file", help="Markdown file to process.", default="dr_egeria_intro_part1.md", required=True,
              prompt=False)
@click.option("--output-folder", help="Output folder.", default="", required=False)
@click.option("--directive", default="process", help="How to process the file",
              type = click.Choice(["display", "validate", "process"], case_sensitive=False), prompt=False, )
@click.option("--server", default=EGERIA_VIEW_SERVER, help="Egeria view server to use.")
@click.option("--url", default=EGERIA_VIEW_SERVER_URL, help="URL of Egeria platform to connect to")
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--user_pass", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--parse-summary", default="none", help="When to show parse summaries",
              type=click.Choice(["all", "errors", "none"], case_sensitive=False))
@click.option("--attribute-logs", default="debug", help="Per-attribute log verbosity",
              type=click.Choice(["debug", "info", "none"], case_sensitive=False))
@click.option("--usage-level", default=None, help="Egeria usage level (Basic or Advanced)",
              type=click.Choice(["Basic", "Advanced"], case_sensitive=False))
@logger.catch
def process_markdown_file(input_file: str, output_folder:str, directive: str, server: str, url: str, userid: str,
                          user_pass: str, parse_summary: str, attribute_logs: str, usage_level: str) -> None:
    """
    Process a markdown file by parsing and executing Dr. Egeria md_commands. Write output to a new file.
    """
    try:
        # Instantiate the client
        from pyegeria import EgeriaTech
        client = EgeriaTech(server, url, userid, user_pass)
        client.create_egeria_bearer_token()
        asyncio.run(process_md_file_v2(
            input_file=input_file, 
            output_folder=output_folder, 
            directive=directive, 
            client=client,
            parse_summary=parse_summary,
            attribute_logs=attribute_logs,
            usage_level=usage_level
        ))
        logger.info(f"Called process_markdown_file with input file {input_file}")
    except PyegeriaException as e:
        logger.error(f"Error processing markdown file {input_file}: {e}")
    except Exception as e:
        logger.error(f"Unexpected error processing markdown file {input_file}: {e}")


def _running_in_pycharm_debugger() -> bool:
    return sys.gettrace() is not None or os.environ.get("PYCHARM_HOSTED") is not None

if __name__ == "__main__":
    if _running_in_pycharm_debugger():
        input_file = Prompt.ask("Markdown File name to process:", default="dr_egeria_intro_part1.md")
        directive = Prompt.ask("How to process the file?", default="validate")
        usage_level = Prompt.ask("Usage level (basic, advanced):", default="basic")
        process_markdown_file.callback(input_file, "", directive,
                              EGERIA_VIEW_SERVER, EGERIA_VIEW_SERVER_URL, EGERIA_USER,
                        EGERIA_USER_PASSWORD, parse_summary="all", attribute_logs="debug", usage_level=usage_level)
    else:
        process_markdown_file()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.



This module is for testing the classification manager view service class and methods.
The routines assume that pytest is being used as the test tool and framework.

A running Egeria environment is needed to run these tests.

"""

import json
import time

from pydantic import ValidationError
from rich import print, print_json, traceback
from rich.console import Console

from pyegeria import EgeriaTech
from pyegeria.core._exceptions import PyegeriaException, print_basic_exception, print_validation_error, \
    PyegeriaConnectionException, PyegeriaAPIException, print_exception_table
from pyegeria.omvs.classification_explorer import ClassificationExplorer

# from pyegeria.output_formatter import make_preamble, make_md_attribute

disable_ssl_warnings = True

platform_url = "https://localhost:9443"
view_server = "qs-view-server"

# c_client = ClassificationExplorer(view_server, platform_url)

user = "erinoverview"
password = "secret"

# element_guid = '4fe24e34-490a-43f0-a0d4-fe45ac45c663'
# element_guid = "a2915132-9d9a-4449-846f-43a871b5a6a0"
# element_guid = "b359e297-a565-414a-8213-fa423312ab36" # clinical trials management
# element_guid = "25b1791f-c2fb-4b93-b236-cad53739a9a2"  # Approved Hospital
# element_guid = "e656f7ca-c11a-4d05-a5ed-adc69abbf0fd"  # a digital product glossary
element_guid = "c28ea54e-060a-4738-9d8c-1b4000442d7e"
relationship_type = "GovernedBy"

# bearer_token = c_client.create_egeria_bearer_token(user, password)
console = Console()


def jprint(info, comment=None):
    if comment:
        print(comment)
    print(json.dumps(info, indent=2))


def valid_guid(guid):
    if (guid is None) or (type(guid) is not str):
        return False
    else:
        return True


#
##
#
def test_get_classified_elements_by():
    # metadata_element_type = 'CertificationType'
    #
    # metadata_element_type = "DeployedDatabaseSchema"
    classification_name = "Confidentiality"
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "LevelIdentifierQueryProperties",
            "levelIdentifier": 1
        }
        response = c_client.get_classified_elements_by(classification_name)

        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is: " + response)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    except ValidationError as e:
        print_validation_error(e)
    except Exception as e:
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_security_tagged_elements():
    # metadata_element_type = 'CertificationType'
    #
    # metadata_element_type = "DeployedDatabaseSchema"
    classification_name = "confidentiality"
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "SecurityTagQueryProperties",

            "metadataElementTypeName": None,
            "limitResultsByStatus": [],
            "sequencingProperty": "displayName",
            "sequencingOrder": "LAST_UPDATE_RECENT",
            "securityLabels": [],
            "securityProperties": {
                "propertyName": "propertyValue"
            },
            "accessGroups": {
                "groupName": []
            }
        }
        response = c_client.get_security_tagged_elements(body, output_format="DICT", report_spec="Referenceable")

        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is: " + response)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    except ValidationError as e:
        print_validation_error(e)
    except Exception as e:
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_owners_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        owner_name = "Erin Overview"
        body = {
            "class": "FilterRequestBody",
            "filter": owner_name
        }
        response = c_client.get_owners_elements(owner_name, body, output_format="DICT", report_spec="Referenceable")

        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is: " + response)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    except ValidationError as e:
        print_validation_error(e)
    except Exception as e:
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_elements_by_origin():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        home_metadata_collection = "PostgresContentPack"
        body = {
            "class": "FindDigitalResourceOriginProperties",
            "metadataElementTypeName": "ValidValueDefinition",
            "otherOriginValues": {"homeMetadataCollectionName": home_metadata_collection}
        }
        response = c_client.get_elements_by_origin(body, output_format="JSON", report_spec="Referenceable")

        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is: " + response)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    except ValidationError as e:
        print_validation_error(e)
    except Exception as e:
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_elements():
    # metadata_element_type = 'CertificationType'
    #
    # metadata_element_type = "DeployedDatabaseSchema"
    open_metadata_type_name = "UserIdentity"
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "ResultsRequestBody",
            "metadataElementTypeName": open_metadata_type_name,
        }
        response = c_client.get_elements(metadata_element_type = open_metadata_type_name,
                                         output_format="DICT",
                                         report_spec="Referenceable",
                                         body=body)

        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is" + response)
        assert True
    except (PyegeriaException, PyegeriaAPIException) as e:
        if e.response_egeria_msg_id == "OMAG-COMMON-400-018":
            print("\n\n==>Invalid OM Type\n\n")
        else:
            print_basic_exception(e)
            assert False, "Invalid request"
    except Exception as e:
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_elements_by_property_value():
    # metadata_element_type = 'Project'
    # property_value = "Campaign:Clinical Trials Management"
    # metadata_element_type = "ValidValueDefinition"
    metadata_element_type_name = None
    # property_value = "Unity Catalog Catalog"
    # property_names = ["name", "qualifiedName"]
    # metadata_element_type = "Asset"
    # property_value = "ClinicalTrials@CocoPharmaceuticals:set-up-clinical-trial"
    # property_value = "default"
    # property_names = ["name", "qualifiedName"]
    property_names = ["name", "displayName", 'qualifiedName']
    property_value = "BusinessArea::RES"
    # property_names = ["anchorGUID"]
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        start_time = time.perf_counter()
        result = c_client.get_elements_by_property_value(
            property_value,
            property_names,
            metadata_element_type=metadata_element_type_name,
            as_of_time="2025-12-01",
        )
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if type(result) is list:
            print(f"\n\tElement count is: {len(result)}")
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is: " + result)

        assert True

    except (
            PyegeriaException
    ) as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_find_elements_by_property_value():
    # metadata_element_type = 'Project'
    # property_value = "Campaign:Clinical Trials Management"
    # metadata_element_type = "ValidValueDefinition"
    # metadata_element_type = None
    # metadata_element_type = "ArchiveFile"
    open_metadata_type_name = "DigitalProduct"
    # metadata_element_type = None
    # property_names = ["name"]
    # property_value = "Set up new clinical trial"
    property_names = ["displayName"]
    property_value = "Open Metadata Digital Product Data Dictionary"

    try:
        c_client = EgeriaTech(view_server, platform_url, user, password)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        start_time = time.perf_counter()
        result = c_client.find_elements_by_property_value(
            property_value, property_names, open_metadata_type_name,
            output_format="JSON", report_spec="Referenceable"
        )
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if type(result) is list:
            print(f"\n\tElement count is: {len(result)}")
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is: " + result)

        assert True

    except (
            PyegeriaException
    ) as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_element_by_guid():
    element_guid = 'dae76285-bef7-45a0-9f32-4854294d969b'
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        start_time = time.perf_counter()
        result = c_client.get_element_by_guid(element_guid, output_format="JSON", report_spec="Referenceable")
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if isinstance(result, list|dict):
            print(f"\n\tElement count is: {len(result)}")
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is: " + result)

        assert True

    except (
            PyegeriaException
    ) as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_actor_for_guid():
    element_guid = "dcfd7e32-8074-4cdf-bdc5-9a6f28818a9d"
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        start_time = time.perf_counter()
        result = c_client.get_actor_for_guid(element_guid)
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if type(result) is dict:
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is: " + result)

        assert True

    except PyegeriaException as e:
        print_basic_exception(e)
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    except ValidationError as e:
        print_validation_error(e)
        console.print_exception(show_locals=True)
    finally:
        c_client.close_session()


def test_get_guid_for_name():
    open_metadata_type_name = None
    property_value = "PostgreSQLServer::CreateAndSurveyGovernanceActionProcess"
    # property_value = "simple-metadata-store"
    # property_value = "Sustainability Glossary"
    # property_value = "qs-view-server"
    # property_value = "Campaign:Sustainability"
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    start_time = time.perf_counter()
    result = c_client.get_guid_for_name(property_value)
    duration = time.perf_counter() - start_time
    print(f"\n\tDuration was {duration} seconds")
    if type(result) is list:
        print(f"\n\tElement count is: {len(result)}")
        print_json(data=result)
    elif type(result) is str:
        console.print("\n\n\t Response is " + result)

    assert True


def test_get_element_guid_by_unique_name():
    open_metadata_type_name = None
    # property_value = "Person:UK:324713"
    # property_value = "simple-metadata-store"
    # property_value = "FileDirectory:CreateAndSurveyGovernanceActionProcess"
    property_value = "Open Metadata Digital Product Data Dictionary"

    c_client = ClassificationExplorer(view_server, platform_url)
    try:
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        start_time = time.perf_counter()
        result = c_client.get_element_guid_by_unique_name(property_value, "displayName")
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if isinstance(result, list|dict):
            print(f"\n\tElement count is: {len(result)}")
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is " + result)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"


def test_get_element_by_unique_name():
    open_metadata_type_name = None
    # property_value = "Person:UK:324713"
    property_value = "FileDirectory:CreateAndSurveyGovernanceActionProcess"

    c_client = ClassificationExplorer(view_server, platform_url)
    try:
        bearer_token = c_client.create_egeria_bearer_token(user, password)

        start_time = time.perf_counter()
        result = c_client.get_element_by_unique_name(property_value, "qualifiedName", output_format="DICT",
                                                     report_spec="Referenceable")
        duration = time.perf_counter() - start_time
        print(f"\n\tDuration was {duration} seconds")
        if isinstance(result, list | dict):
            print(f"\n\tElement count is: {len(result)}")
            print(json.dumps(result, indent=4))
        elif type(result) is str:
            console.print("\n\n\t Response is " + result)

        assert True
    except (PyegeriaException, PyegeriaConnectionException) as e:
        print_basic_exception(e)
        assert False


def test_get_elements_by_classification():
    # metadata_element_type = "Project"
    # metadata_element_type = "DeployedDatabaseSchema"
    open_metadata_type_name = None
    # classification = "GovernanceProject"
    classification = "Ownership"
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    body = {
        "class": "ResultsRequestBody",
        "metadataElementTypeName": open_metadata_type_name,
        "pageSize": 10,
    }
    response = c_client.get_elements_by_classification(
        classification, output_format="JSON", report_spec="Collections", body=body
    )

    if type(response) is list:
        print("Result = \n")
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is: " + response)

    assert True


def test_get_elements_by_classification_with_property_value():
    # metadata_element_type = "Project"
    open_metadata_type_name = None
    classification = "DataSpec"
    # property_value = "Collection"
    # property_names = ["anchorTypeName"]
    property_value = ""
    property_names = [""]
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        result = c_client.get_elements_by_classification_with_property_value(
            classification,
            property_value,
            property_names,
            metadata_element_type_name=open_metadata_type_name,
        )

        if type(result) is list:
            print_json(data=result)
        elif type(result) is str:
            console.print("\n\n\t Response is: " + result)

        assert True

    except (PyegeriaException) as e:
        print_basic_exception(e)
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_find_elements_by_classification_with_property_value():
    # classification = "GovernanceProject"
    # metadata_element_type = "Project"
    # property_value = "Clinical Trials"
    # property_names = ["name", "qualifiedName"]
    #
    classification = "Anchor"
    # metadata_element_type = "Project"
    open_metadata_type_name = "DigitalProduct"
    property_value = "Open Metadata Digital Product Data Dictionary"
    property_names = ["displayName"]
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)

        start_time = time.perf_counter()
        response = c_client.find_elements_by_classification_with_property_value(
            classification, property_value, property_names, open_metadata_type_name
        )
        duration = time.perf_counter() - start_time
        print(
            f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
        )
        if type(response) is list:
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is: " + response)

        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False


def test_find_anchored_elements_with_property_value():
    classification = "Anchors"
    open_metadata_type_name = None
    property_value = "PostgreSQL Server"
    # property_names = ["ServerCapability", "anchorTypeName"]
    property_names = ["name"]
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    response = c_client.find_elements_by_classification_with_property_value(
        classification, property_value, property_names, open_metadata_type_name
    )

    if type(response) is list:
        print("Response payload is: \n")
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is" + response)

    assert True


def test_get_all_related_elements():
    # metadata_element_type = 'Project'
    open_metadata_type_name = None
    c_client = ClassificationExplorer(view_server, platform_url)
    # element_guid = "d156faa6-90cf-4be8-b3c1-c002f3e9a0e5" # branch database
    # element_guid = "da0442bf-818f-406b-99dc-83b72605cc98"
    # element_guid = "8b9cce34-ff42-4f9d-b4b3-6317c8a767c3"  # Retail schema
    element_guid = "0182bacf-32c4-40f0-91b5-2462dfeab50c"
    bearer_token = c_client.create_egeria_bearer_token(user, password)
    response = c_client.get_related_elements(element_guid, relationship_type=None)

    if type(response) is list:
        print(f"\n\tElement count is: {len(response)}")
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is" + response)

    assert True


def test_get_related_elements():
    # metadata_element_type = 'CertificationType'
    # element_guid = '2ce8d10d-08a2-4fca-b0c2-1d5a335d00fc'
    element_guid = "10662294-52d0-43c5-9aa3-19922e478e69"
    # metadata_element_type = "Organization"
    # metadata_element_type = "CSVFile"
    # metadata_element_type = "InformationSupplyChain"
    open_metadata_type_name = "Referenceable"
    # element_guid = "8dca6e76-d454-4344-9c93-faa837a1a898"
    # relationship_type = "DataContentForDataSet"
    relationship_type = "CollectionMembership"
    # relationship_type = "InformationSupplyChainComposition"
    c_client = ClassificationExplorer(view_server, platform_url)
    try:
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "ResultsRequestBody",
            "metadataElementTypeName": open_metadata_type_name,
        }
        response = c_client.get_related_elements(
            element_guid, relationship_type, body=body
        )

        if type(response) is list:
            print_json(data=response)
        elif type(response) is str:
            console.print("\n\n\t Response is:\n " + response)

        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
        assert False, "Invalid request"


def test_get_related_elements_with_property_value():
    # metadata_element_type = 'Project'
    open_metadata_type_name = None
    relationship_type = "ResourceList"
    property_value = "Catalog Resource"
    property_names = [
        "resourceUse",
    ]
    element_guid = '2ce8d10d-08a2-4fca-b0c2-1d5a335d00fc'
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        result = c_client.get_related_elements_with_property_value(
            element_guid,
            relationship_type,
            property_value,
            property_names,
            metadata_element_type_name=open_metadata_type_name,
        )

        if type(result) is list:
            print_json(data=result)
        elif type(result) is str:
            print("\n\n\t Response is: " + result)
        else:
            print(f"type is: {type(result)}")

        assert True

    except (
            PyegeriaException) as e:
        print_basic_exception(e)
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_find_related_elements_with_property_value():
    # metadata_element_type = 'Project'
    open_metadata_type_name = None
    property_value = "Clinical Trials Management"
    property_names = ["name", "qualifiedName"]
    # property_value = "Partner"
    # property_names = ["teamType"]

    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    response = c_client.find_related_elements_with_property_value(
        element_guid,
        relationship_type,
        property_value,
        property_names,
        open_metadata_type_name,
    )

    if type(response) is list:
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is " + response)

    assert True

def test_find_root_elements():
    # metadata_element_type = 'Project'
    open_metadata_type_name = "ValidMetadataValue"
    property_value = "Clinical Trials Management"
    property_names = ["name", "qualifiedName"]
    # property_value = "Partner"
    # property_names = ["teamType"]
    search_properties =  {
            "class" : "SearchProperties",
            "conditions": [ {
              "property" : "preferredValue",
              "operator": "IS_NULL"}],
            "matchCriteria": "ANY"
          }

    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    response = c_client.find_root_elements(
        open_metadata_type_name,
        search_properties

    )

    if type(response) is list:
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is " + response)

    assert True


def test_get_relationships():
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    relationship_type = "Certification"
    response = c_client.get_relationships(relationship_type, output_format="JSON")

    if type(response) is list:
        print(f"\n\tElement count is: {len(response)}")
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is: " + response)

    assert True


def test_get_relationships_with_property_value():
    property_value = "Organization:Hampton Hospital"
    property_names = ["name", "qualifiedName"]
    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        result = c_client.get_relationships_with_property_value(
            property_value,
            property_names,
            relationship_type=relationship_type,
            output_format="JSON",
        )

        if type(result) is list:
            print_json(data=result)
        elif type(result) is str:
            print("\n\n\t Response is: " + result)
        else:
            print(f"type is: {type(result)}")

        assert True

    except (
            PyegeriaException) as e:
        print_basic_exception(e)
        console.print_exception(show_locals=True)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_find_relationships_with_property_value():
    property_value = "Clinical Trials"
    property_names = ["name", "qualifiedName"]
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    response = c_client.find_relationships_with_property_value(
        property_value,
        property_names,
        relationship_type=relationship_type,
    )

    if type(response) is list:
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is " + response)

    assert True


def test_retrieve_instance_for_guid():
    c_client = ClassificationExplorer(view_server, platform_url)

    bearer_token = c_client.create_egeria_bearer_token(user, password)
    element_guid = "59f0232c-f834-4365-8e06-83695d238d2d"
    response = c_client.retrieve_instance_for_guid(element_guid, output_format="JSON")

    if type(response) is dict:
        print_json(data=response)
    elif type(response) is str:
        console.print("\n\n\t Response is " + response)

    assert True


def test_set_criticality_classification():
    # metadata_element_type = 'Project'
    element_guid = "9f6f668d-d9b8-44e3-915b-9edeb4c1f8a5"
    body = {
        "class": "NewClassificationRequestBody",
        "Properties": {
            "class": "CriticalityProperties",
            "levelIdentifier": 3,
            "criticality": 3
        }
    }

    try:
        c_client = ClassificationExplorer(view_server, platform_url)

        bearer_token = c_client.create_egeria_bearer_token(user, password)
        result = c_client.set_criticality_classification(
            element_guid,
            body
        )

        if type(result) is list:
            print_json(data=result)
        elif type(result) is str:
            print("\n\n\t Response is: " + result)
        else:
            print(f"type is: {type(result)}")

        assert True

    except (
            PyegeriaException
    ) as e:
        print_basic_exception(e)
        assert False, "Invalid request"
    finally:
        c_client.close_session()


def test_get_subject_area_members():
    subject_area = "Clinical Trials"
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "FilterRequestBody",
            "filter": "*",
        }
        response = c_client.get_subject_area_members(subject_area, body)
        if type(response) is list:
            print(f"\n\tElement count is: {len(response)}")
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_meanings():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "FilterRequestBody",
            "filter": "*",
        }
        response = c_client.get_meanings(element_guid, body)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_semantic_asignees():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        body = {
            "class": "FilterRequestBody",
            "filter": "*",
        }
        response = c_client.get_semantic_asignees(element_guid, body)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_governed_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_governed_elements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_governed_by_definitions():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_governed_by_definitions(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_source_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_source_elements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_elements_sourced_from():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_elements_sourced_from(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_scopes():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_scopes(element_guid)
        if isinstance(response, list|dict):
            print(f"\n\tElement count is: {len(response)}")
            print(json.dumps(response, indent=2))
            assert True
        else:
            print(f"Expected list, got {type(response).__name__}, {response}")
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_scoped_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_scoped_elements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_licensed_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_licensed_elements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_licenses():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_licenses(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_certified_elements():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_certified_elements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_get_certifications():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        response = c_client.get_certifications(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_confidence_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "ConfidenceProperties",
            "confidence": 100,
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_confidence_classification(element_guid, body)
        c_client.clear_confidence_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_confidentiality_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "ConfidentialityProperties",
            "levelIdentifier": 1,
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_confidentiality_classification(element_guid, body)
        c_client.clear_confidentiality_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_impact_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "ImpactProperties",
            "levelIdentifier": 1,
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_impact_classification(element_guid, body)
        c_client.clear_impact_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_criticality_classification_clear():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.clear_criticality_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_gov_definition_to_element():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        # using dummy GUIDs for now, expecting potential 404 but testing call pattern
        def_guid = "dummy-def-guid"
        c_client.add_gov_definition_to_element(def_guid, element_guid)
        c_client.remove_gov_definition_from_element(def_guid, element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_scope_to_element():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        scope_guid = "17f9bcd9-ede6-431e-9cec-1b8474f31e4b"

        c_client.add_scope_to_element(scope_guid, element_guid)
        c_client.clear_scope_from_element(scope_guid, element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_assign_actor_to_element():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        actor_guid = "dummy-actor-guid"

        c_client.assign_actor_to_element(element_guid, actor_guid)
        c_client.unassign_actor_from_element(element_guid, actor_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_certification_to_element():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        cert_type_guid = "dummy-cert-type-guid"
        # add returns guid of relationship
        # c_client.add_certification_to_element(cert_type_guid, element_guid)
        # c_client.update_certification("dummy-guid", {})
        # c_client.decertify_element("dummy-guid")
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_license_to_element():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        license_type_guid = "dummy-license-type-guid"
        # c_client.add_license_to_element(license_type_guid, element_guid)
        # c_client.update_license("dummy-guid", {})
        # c_client.unlicense_element("dummy-guid")
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_add_ownership_to_element():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "OwnershipProperties",
            "owner": "72b2af0a-cdd6-477a-b885-08d9f2e1b3b0",
            "ownerType": "USER_ID"
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_ownership_to_element(element_guid, body)
        c_client.clear_ownership_from_element(element_guid)
        assert True
    except PyegeriaException as e:
        print_exception_table(e)
    except Exception as e:
        console.print_exception(show_locals=True)
    finally:
        c_client.close_session()


def test_digital_resource_origin():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "DigitalResourceOriginProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_digital_resource_origin(element_guid, body)
        # c_client.clear_digital_resource_origin_from_element(element_guid, {"class": "DeleteClassificationRequestBody"})
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_zone_membership():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "ZoneMembershipProperties",
            "zoneMembership": ["quarantine"],
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_zone_membership(element_guid, body)
        # c_client.clear_zone_membership(element_guid, {"class": "DeleteClassificationRequestBody"})
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_retention_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "RetentionProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_retention_classification(element_guid, body)
        c_client.clear_retention_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_governance_expectation():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "GovernanceExpectationProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_governance_expectation(element_guid, body)
        c_client.update_governance_expectation(element_guid, body)
        # c_client.clear_governance_expectation(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_security_tags_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "SecurityTagsProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_security_tags_classification(element_guid, body)
        c_client.clear_security_tags_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_semantic_assignment():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        term_guid = "dummy-term-guid"
        # c_client.setup_semantic_assignment(term_guid, element_guid, {})
        # c_client.clear_semantic_assignment_classification(term_guid, element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_subject_area_membership():
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_element_to_subject_area(element_guid, {})
        c_client.remove_element_from_subject_area(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_governance_measurements():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "GovernanceMeasurementsProperties",
        }
    }
    update_body = {
        "class": "UpdateClassificationRequestBody",
        "mergeUpdate": True,
        "properties": {
            "class": "GovernanceMeasurementsProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_governance_measurements(element_guid, body)
        c_client.update_governance_measurements(element_guid, update_body)
        c_client.clear_governance_measurements(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_data_scope_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "DataScopeProperties",
        }
    }
    update_body = {
        "class": "UpdateClassificationRequestBody",
        "mergeUpdate": True,
        "properties": {
            "class": "DataScopeProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.add_data_scope(element_guid, body)
        c_client.update_data_scope(element_guid, update_body)
        c_client.clear_data_scope(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_search_keywords():
    body = {
        "class": "NewAttachmentRequestBody",
        "properties": {
            "class": "SearchKeywordProperties",
            "displayName": "testKeyword",
        }
    }
    update_body = {
        "class": "UpdateElementRequestBody",
        "mergeUpdate": True,
        "properties": {
            "class": "SearchKeywordProperties",
            "displayName": "updatedKeyword",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        keyword_guid = c_client.add_search_keyword_to_element(element_guid, body)
        if keyword_guid != "NO_GUID_RETURNED":
            c_client.update_search_keyword(keyword_guid, update_body)
            c_client.remove_search_keyword_from_element(keyword_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_known_duplicate_classification():
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "KnownDuplicateProperties"
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_known_duplicate_classification(element_guid, body)
        c_client.clear_known_duplicate_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_peer_duplicate_link():
    peer_guid = "peer-guid"
    body = {
        "class": "NewRelationshipRequestBody",
        "properties": {
            "class": "PeerDuplicateLinkProperties",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.link_elements_as_peer_duplicates(element_guid, peer_guid, body)
        c_client.unlink_elements_as_peer_duplicates(element_guid, peer_guid, body)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_consolidated_duplicate_management():
    source_guid = "source-guid"
    body = {
        "class": "NewClassificationRequestBody",
        "properties": {
            "class": "ConsolidatedDuplicateProperties",
        }
    }
    link_body = {
        "class": "NewRelationshipRequestBody",
        "properties": {
            "class": "ConsolidatedDuplicateLinkProperties"
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.set_consolidated_duplicate_classification(element_guid, body)
        c_client.link_consolidated_duplicate_to_source(element_guid, source_guid, link_body)
        c_client.unlink_consolidated_duplicate_from_source_element(element_guid, source_guid, link_body)
        c_client.clear_consolidated_duplicate_classification(element_guid)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()


def test_action_actor_linking():
    action_guid = "action-guid"
    actor_guid = "actor-guid"
    body = {
        "class": "NewRelationshipRequestBody",
        "properties": {
            "class": "AssignmentScopeProperties",
            "label": "testAction",
        }
    }
    try:
        c_client = ClassificationExplorer(view_server, platform_url)
        bearer_token = c_client.create_egeria_bearer_token(user, password)
        c_client.assign_action(action_guid, actor_guid, body)
        c_client.reassign_action(action_guid, actor_guid, body)
        c_client.unassign_action(action_guid, actor_guid, body)
        assert True
    except PyegeriaException as e:
        print_basic_exception(e)
    finally:
        c_client.close_session()



"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module provides access to the Notification Manager OMVS module.

The Notification Manager OMVS provides APIs for managing notification types and their
relationships to monitored resources and subscribers.
"""

import asyncio
from typing import Optional

from pyegeria.core._server_client import ServerClient
from pyegeria.models import (
    NewRelationshipRequestBody,
    DeleteRelationshipRequestBody,
)
from pyegeria.core.utils import dynamic_catch, body_slimmer


class NotificationManager(ServerClient):
    """
    Manage notification types and their relationships to monitored resources and subscribers.

    This client provides methods to link notification types with monitored resources
    and notification subscribers.

    Attributes
    ----------
    view_server : str
        The name of the View Server to connect to.
    platform_url : str
        URL of the server platform to connect to.
    user_id : str
        The identity of the user calling the method - this sets a default optionally
        used by the methods when the user doesn't pass the user_id on a method call.
    user_pwd : str, optional
        The password associated with the user_id. Defaults to None.
    token : str, optional
        An optional bearer token for authentication.

    Methods
    -------
    link_monitored_resource(notification_type_guid, monitored_resource_guid, body)
        Attach a monitored resource to a notification type.
    detach_monitored_resource(notification_type_guid, monitored_resource_guid, body)
        Detach a monitored resource from a notification type.
    link_notification_subscriber(notification_type_guid, subscriber_guid, body)
        Attach a subscriber to a notification type.
    detach_notification_subscriber(notification_type_guid, subscriber_guid, body)
        Detach a subscriber from a notification type.
    """

    def __init__(
        self,
        view_server: str = None,
        platform_url: str = None,
        user_id: str = None,
        user_pwd: Optional[str] = None,
        token: Optional[str] = None,
    ):
        ServerClient.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        self.view_server = self.server_name
        self.platform_url = self.platform_url
        self.user_id = self.user_id
        self.user_pwd = self.user_pwd
        self.notification_manager_command_root: str = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/notification-manager"
        )

    def _prepare_body(self, body: Optional[dict | NewRelationshipRequestBody | DeleteRelationshipRequestBody]) -> dict:
        """Convert Pydantic models to dict and slim the body."""
        if body is None:
            return None
        if isinstance(body, dict):
            return body_slimmer(body)
        # It's a Pydantic model
        return body_slimmer(body.model_dump(mode='json', by_alias=True, exclude_none=True))

    #
    # Monitored Resource Management
    #

    @dynamic_catch
    async def _async_link_monitored_resource(
        self,
        notification_type_guid: str,
        monitored_resource_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a monitored resource to a notification type. Async version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        monitored_resource_guid : str
            The GUID of the monitored resource.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to NewRelationshipRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = (
            f"{self.notification_manager_command_root}/notification-types/"
            f"{notification_type_guid}/monitored-resources/{monitored_resource_guid}/attach"
        )
        await self._async_make_request("POST", url, self._prepare_body(body))

    def link_monitored_resource(
        self,
        notification_type_guid: str,
        monitored_resource_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a monitored resource to a notification type. Sync version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        monitored_resource_guid : str
            The GUID of the monitored resource.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_monitored_resource(
                notification_type_guid, monitored_resource_guid, body
            )
        )

    @dynamic_catch
    async def _async_detach_monitored_resource(
        self,
        notification_type_guid: str,
        monitored_resource_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a monitored resource from a notification type. Async version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        monitored_resource_guid : str
            The GUID of the monitored resource to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = (
            f"{self.notification_manager_command_root}/notification-types/"
            f"{notification_type_guid}/monitored-resources/{monitored_resource_guid}/detach"
        )
        body = self._prepare_body(body)
        if body:
            await self._async_make_request("POST", url, body)
        else:
            await self._async_make_request("POST", url)


    def detach_monitored_resource(
        self,
        notification_type_guid: str,
        monitored_resource_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a monitored resource from a notification type. Sync version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        monitored_resource_guid : str
            The GUID of the monitored resource to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_monitored_resource(
                notification_type_guid, monitored_resource_guid, body
            )
        )

    #
    # Notification Subscriber Management
    #

    @dynamic_catch
    async def _async_link_notification_subscriber(
        self,
        notification_type_guid: str,
        subscriber_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a subscriber to a notification type. Async version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        subscriber_guid : str
            The GUID of the subscriber.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = (
            f"{self.notification_manager_command_root}/notification-types/"
            f"{notification_type_guid}/notification-subscribers/{subscriber_guid}/attach"
        )
        await self._async_make_request("POST", url, self._prepare_body(body))

    def link_notification_subscriber(
        self,
        notification_type_guid: str,
        subscriber_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a subscriber to a notification type. Sync version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        subscriber_guid : str
            The GUID of the subscriber.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_notification_subscriber(
                notification_type_guid, subscriber_guid, body
            )
        )

    @dynamic_catch
    async def _async_detach_notification_subscriber(
        self,
        notification_type_guid: str,
        subscriber_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a subscriber from a notification type. Async version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        subscriber_guid : str
            The GUID of the subscriber to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = (
            f"{self.notification_manager_command_root}/notification-types/"
            f"{notification_type_guid}/notification-subscribers/{subscriber_guid}/detach"
        )
        await self._async_make_request("POST", url, self._prepare_body(body))

    def detach_notification_subscriber(
        self,
        notification_type_guid: str,
        subscriber_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a subscriber from a notification type. Sync version.

        Parameters
        ----------
        notification_type_guid : str
            The GUID of the notification type.
        subscriber_guid : str
            The GUID of the subscriber to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_notification_subscriber(
                notification_type_guid, subscriber_guid, body
            )
        )